from dataclasses import dataclass
from os import getenv
from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Type, Union

from pydantic import BaseModel

from agno.exceptions import ModelProviderError
from agno.models.base import Model
from agno.models.message import Message
from agno.models.metrics import MessageMetrics
from agno.models.response import ModelResponse
from agno.run.agent import RunOutput
from agno.utils.log import log_debug, log_error, log_warning
from agno.utils.models.watsonx import format_images_for_message

try:
    from ibm_watsonx_ai import Credentials
    from ibm_watsonx_ai.foundation_models import ModelInference
except ImportError:
    raise ImportError("`ibm-watsonx-ai` is not installed. Please install it using `pip install ibm-watsonx-ai`.")


@dataclass
class WatsonX(Model):
    """
    A class for interacting with IBM WatsonX models.

    See supported models at: https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-models.html?context=wx

    For more information, see: https://www.ibm.com/watsonx/developer/
    """

    id: str = "ibm/granite-20b-code-instruct"
    name: str = "WatsonX"
    provider: str = "IBM"

    # Request parameters
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    logprobs: Optional[int] = None
    top_logprobs: Optional[int] = None

    request_params: Optional[Dict[str, Any]] = None

    # Client parameters
    api_key: Optional[str] = None
    project_id: Optional[str] = None
    url: Optional[str] = "https://eu-de.ml.cloud.ibm.com"
    verify: bool = True
    client_params: Optional[Dict[str, Any]] = None

    # WatsonX client
    model_client: Optional[ModelInference] = None

    def _get_client_params(self) -> Dict[str, Any]:
        # Fetch API key and project ID from env if not already set
        self.api_key = self.api_key or getenv("IBM_WATSONX_API_KEY")
        if not self.api_key:
            log_error("IBM_WATSONX_API_KEY not set. Please set the IBM_WATSONX_API_KEY environment variable.")

        self.project_id = self.project_id or getenv("IBM_WATSONX_PROJECT_ID")
        if not self.project_id:
            log_error("IBM_WATSONX_PROJECT_ID not set. Please set the IBM_WATSONX_PROJECT_ID environment variable.")

        self.url = getenv("IBM_WATSONX_URL") or self.url

        # Create credentials object
        credentials = Credentials(url=self.url, api_key=self.api_key, verify=self.verify)

        return {
            "credentials": credentials,
            "project_id": self.project_id,
        }

    def get_client(self) -> ModelInference:
        """
        Returns a WatsonX ModelInference client.

        Returns:
            ModelInference: An instance of the WatsonX ModelInference client.
        """
        if self.model_client:
            return self.model_client

        # Get client parameters
        client_params = self._get_client_params()

        # Initialize model inference client
        self.model_client = ModelInference(model_id=self.id, **client_params)

        return self.model_client

    def get_request_params(
        self,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> Dict[str, Any]:
        params = {
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "logprobs": self.logprobs,
            "top_logprobs": self.top_logprobs,
            "response_format": response_format,
        }
        # Filter out None values
        params = {k: v for k, v in params.items() if v is not None}
        request_params = {}

        if params:
            request_params["params"] = params

        # Add tools
        if tools is not None:
            request_params["tools"] = tools  # type: ignore
            if tool_choice is not None:
                request_params["tool_choice"] = tool_choice  # type: ignore
        # Add additional request params if provided
        if self.request_params:
            request_params.update(self.request_params)

        if request_params:
            log_debug(f"Calling {self.provider} with request parameters: {request_params}", log_level=2)
        return request_params

    def _format_message(self, message: Message, compress_tool_results: bool = False) -> Dict[str, Any]:
        """
        Format a message into the format expected by WatsonX.

        Args:
            message (Message): The message to format.
            compress_tool_results: Whether to compress tool results.

        Returns:
            Dict[str, Any]: The formatted message.
        """
        if message.images is not None and isinstance(message.content, str):
            message = format_images_for_message(message=message, images=message.images)

        if message.audio is not None and len(message.audio) > 0:
            log_warning("Audio input is currently unsupported.")

        if message.files is not None and len(message.files) > 0:
            log_warning("File input is currently unsupported.")

        if message.videos is not None and len(message.videos) > 0:
            log_warning("Video input is currently unsupported.")

        message_dict = message.to_dict()

        # Use compressed content for tool messages if compression is active
        if message.role == "tool" and compress_tool_results:
            message_dict["content"] = message.get_content(use_compressed_content=True)
        return message_dict

    def invoke(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> ModelResponse:
        """
        Send a chat completion request to the WatsonX API.
        """
        try:
            client = self.get_client()

            formatted_messages = [self._format_message(m, compress_tool_results) for m in messages]
            request_params = self.get_request_params(
                response_format=response_format, tools=tools, tool_choice=tool_choice
            )

            assistant_message.metrics.start_timer()
            response = client.chat(messages=formatted_messages, **request_params)
            assistant_message.metrics.stop_timer()

            model_response = self._parse_provider_response(response, response_format=response_format)

            return model_response

        except Exception as e:
            log_error(f"Error calling WatsonX API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    async def ainvoke(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> Any:
        """
        Sends an asynchronous chat completion request to the WatsonX API.
        """
        try:
            client = self.get_client()
            formatted_messages = [self._format_message(m, compress_tool_results) for m in messages]

            request_params = self.get_request_params(
                response_format=response_format, tools=tools, tool_choice=tool_choice
            )

            assistant_message.metrics.start_timer()
            provider_response = await client.achat(messages=formatted_messages, **request_params)
            assistant_message.metrics.stop_timer()

            model_response = self._parse_provider_response(provider_response, response_format=response_format)

            return model_response

        except Exception as e:
            log_error(f"Error calling WatsonX API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    def invoke_stream(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> Iterator[ModelResponse]:
        """
        Send a streaming chat completion request to the WatsonX API.
        """
        try:
            client = self.get_client()
            formatted_messages = [self._format_message(m, compress_tool_results) for m in messages]

            request_params = self.get_request_params(
                response_format=response_format, tools=tools, tool_choice=tool_choice
            )

            assistant_message.metrics.start_timer()

            for chunk in client.chat_stream(messages=formatted_messages, **request_params):
                yield self._parse_provider_response_delta(chunk)

            assistant_message.metrics.stop_timer()

        except Exception as e:
            log_error(f"Error calling WatsonX API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    async def ainvoke_stream(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> AsyncIterator[ModelResponse]:
        """
        Sends an asynchronous streaming chat completion request to the WatsonX API.
        """
        try:
            client = self.get_client()
            formatted_messages = [self._format_message(m, compress_tool_results) for m in messages]

            # Get parameters for chat
            request_params = self.get_request_params(
                response_format=response_format, tools=tools, tool_choice=tool_choice
            )

            assistant_message.metrics.start_timer()

            async_stream = await client.achat_stream(messages=formatted_messages, **request_params)
            async for chunk in async_stream:
                yield self._parse_provider_response_delta(chunk)

            assistant_message.metrics.stop_timer()

        except Exception as e:
            log_error(f"Error in async streaming from WatsonX API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    # Override base method
    @staticmethod
    def parse_tool_calls(tool_calls_data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Build tool calls from streamed tool call data.

        Args:
            tool_calls_data (List[ChoiceDeltaToolCall]): The tool call data to build from.

        Returns:
            List[Dict[str, Any]]: The built tool calls.
        """
        tool_calls: List[Dict[str, Any]] = []
        for _tool_call in tool_calls_data:
            _index = _tool_call.get("index", 0)
            _tool_call_id = _tool_call.get("id")
            _tool_call_type = _tool_call.get("type")
            _function_name = _tool_call.get("function", {}).get("name")
            _function_arguments = _tool_call.get("function", {}).get("arguments")

            if len(tool_calls) <= _index:
                tool_calls.extend([{}] * (_index - len(tool_calls) + 1))
            tool_call_entry = tool_calls[_index]
            if not tool_call_entry:
                tool_call_entry["id"] = _tool_call_id
                tool_call_entry["type"] = _tool_call_type
                tool_call_entry["function"] = {
                    "name": _function_name or "",
                    "arguments": _function_arguments or "",
                }
            else:
                if _function_name:
                    tool_call_entry["function"]["name"] += _function_name
                if _function_arguments:
                    tool_call_entry["function"]["arguments"] += _function_arguments
                if _tool_call_id:
                    tool_call_entry["id"] = _tool_call_id
                if _tool_call_type:
                    tool_call_entry["type"] = _tool_call_type
        return tool_calls

    def _parse_provider_response(
        self,
        response: Dict[str, Any],
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
    ) -> ModelResponse:
        """
        Parse the WatsonX response into a ModelResponse.
        """
        model_response = ModelResponse()

        # Get response message
        response_message = response["choices"][0]["message"]
        # Parse structured outputs if enabled
        try:
            if (
                response_format is not None
                and isinstance(response_format, type)
                and issubclass(response_format, BaseModel)
            ):
                parsed_object = response_message.parsed  # type: ignore
                if parsed_object is not None:
                    model_response.parsed = parsed_object
        except Exception as e:
            log_warning(f"Error retrieving structured outputs: {e}")

        # Add role
        if response_message.get("role") is not None:
            model_response.role = response_message["role"]

        # Add content
        if response_message.get("content") is not None:
            model_response.content = response_message["content"]

        # Add tool calls
        if response_message.get("tool_calls") is not None and len(response_message["tool_calls"]) > 0:
            try:
                model_response.tool_calls = response_message["tool_calls"]
            except Exception as e:
                log_warning(f"Error processing tool calls: {e}")

        if response.get("usage") is not None:
            model_response.response_usage = self._get_metrics(response["usage"])

        return model_response

    def _parse_provider_response_delta(self, response_delta: Dict[str, Any]) -> ModelResponse:
        """
        Parse the OpenAI streaming response into a ModelResponse.
        """
        model_response = ModelResponse()

        if response_delta.get("choices") and len(response_delta["choices"]) > 0:
            choice_delta: Dict[str, Any] = response_delta["choices"][0].get("delta")

            if choice_delta:
                # Add content
                if choice_delta.get("content") is not None:
                    model_response.content = choice_delta["content"]

                # Add tool calls
                if choice_delta.get("tool_calls") is not None:
                    model_response.tool_calls = choice_delta["tool_calls"]

        # Add usage metrics if present
        if response_delta.get("usage") is not None:
            model_response.response_usage = self._get_metrics(response_delta["usage"])

        return model_response

    def _get_metrics(self, response_usage: Dict[str, Any]) -> MessageMetrics:
        """
        Parse the given WatsonX usage into an Agno MessageMetrics object.

        Args:
            response_usage: Usage data from WatsonX

        Returns:
            MessageMetrics: Parsed metrics data
        """
        metrics = MessageMetrics()

        metrics.input_tokens = response_usage.get("prompt_tokens") or 0
        metrics.output_tokens = response_usage.get("completion_tokens") or 0
        metrics.total_tokens = metrics.input_tokens + metrics.output_tokens

        return metrics



from agno.vectordb.weaviate.index import Distance, VectorIndex
from agno.vectordb.weaviate.weaviate import SearchType, Weaviate

__all__ = [
    "Distance",
    "VectorIndex",
    "Weaviate",
    "SearchType",
]



"""Approval API router -- list, resolve, and delete human approvals."""

import asyncio
import time
from typing import Any, Dict, Literal, Optional

from fastapi import APIRouter, Depends, HTTPException, Query

from agno.os.routers.approvals.schema import (
    ApprovalCountResponse,
    ApprovalResolve,
    ApprovalResponse,
    ApprovalStatusResponse,
)
from agno.os.schema import PaginatedResponse, PaginationInfo


def get_approval_router(os_db: Any, settings: Any) -> APIRouter:
    """Factory that creates and returns the approval router.

    Args:
        os_db: The AgentOS-level DB adapter (must support approval methods).
        settings: AgnoAPISettings instance.

    Returns:
        An APIRouter with all approval endpoints attached.
    """
    from agno.os.auth import get_authentication_dependency

    router = APIRouter(tags=["Approvals"])
    auth_dependency = get_authentication_dependency(settings)

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------

    async def _db_call(method_name: str, *args: Any, **kwargs: Any) -> Any:
        fn = getattr(os_db, method_name, None)
        if fn is None:
            raise HTTPException(status_code=503, detail="Approvals not supported by the configured database")
        try:
            if asyncio.iscoroutinefunction(fn):
                return await fn(*args, **kwargs)
            return fn(*args, **kwargs)
        except NotImplementedError:
            raise HTTPException(status_code=503, detail="Approvals not supported by the configured database")

    # ------------------------------------------------------------------
    # Endpoints
    # ------------------------------------------------------------------

    @router.get("/approvals", response_model=PaginatedResponse[ApprovalResponse])
    async def list_approvals(
        status: Optional[Literal["pending", "approved", "rejected", "expired", "cancelled"]] = Query(None),
        source_type: Optional[str] = Query(None),
        approval_type: Optional[Literal["required", "audit"]] = Query(None),
        pause_type: Optional[str] = Query(None),
        agent_id: Optional[str] = Query(None),
        team_id: Optional[str] = Query(None),
        workflow_id: Optional[str] = Query(None),
        user_id: Optional[str] = Query(None),
        schedule_id: Optional[str] = Query(None),
        run_id: Optional[str] = Query(None),
        limit: int = Query(100, ge=1, le=1000),
        page: int = Query(1, ge=1),
        _: bool = Depends(auth_dependency),
    ) -> PaginatedResponse[ApprovalResponse]:
        approvals, total_count = await _db_call(
            "get_approvals",
            status=status,
            source_type=source_type,
            approval_type=approval_type,
            pause_type=pause_type,
            agent_id=agent_id,
            team_id=team_id,
            workflow_id=workflow_id,
            user_id=user_id,
            schedule_id=schedule_id,
            run_id=run_id,
            limit=limit,
            page=page,
        )
        total_pages = (total_count + limit - 1) // limit if total_count > 0 else 0
        return PaginatedResponse(
            data=approvals,
            meta=PaginationInfo(
                page=page,
                limit=limit,
                total_pages=total_pages,
                total_count=total_count,
            ),
        )

    @router.get("/approvals/count", response_model=ApprovalCountResponse)
    async def get_approval_count(
        user_id: Optional[str] = Query(None),
        _: bool = Depends(auth_dependency),
    ) -> Dict[str, int]:
        count = await _db_call("get_pending_approval_count", user_id=user_id)
        return {"count": count}

    @router.get("/approvals/{approval_id}/status", response_model=ApprovalStatusResponse)
    async def get_approval_status(
        approval_id: str,
        _: bool = Depends(auth_dependency),
    ) -> ApprovalStatusResponse:
        approval = await _db_call("get_approval", approval_id)
        if approval is None:
            raise HTTPException(status_code=404, detail="Approval not found")
        return ApprovalStatusResponse(
            approval_id=approval_id,
            status=approval.get("status", "unknown"),
            run_id=approval.get("run_id", ""),
            resolved_at=approval.get("resolved_at"),
            resolved_by=approval.get("resolved_by"),
        )

    @router.get("/approvals/{approval_id}", response_model=ApprovalResponse)
    async def get_approval(
        approval_id: str,
        _: bool = Depends(auth_dependency),
    ) -> Dict[str, Any]:
        approval = await _db_call("get_approval", approval_id)
        if approval is None:
            raise HTTPException(status_code=404, detail="Approval not found")
        return approval

    @router.post("/approvals/{approval_id}/resolve", response_model=ApprovalResponse)
    async def resolve_approval(
        approval_id: str,
        body: ApprovalResolve,
        _: bool = Depends(auth_dependency),
    ) -> Dict[str, Any]:
        now = int(time.time())
        update_kwargs: Dict[str, Any] = {
            "status": body.status,
            "resolved_by": body.resolved_by,
            "resolved_at": now,
        }
        if body.resolution_data is not None:
            update_kwargs["resolution_data"] = body.resolution_data
        result = await _db_call(
            "update_approval",
            approval_id,
            expected_status="pending",
            **update_kwargs,
        )
        if result is None:
            # Either the approval doesn't exist or it was already resolved
            existing = await _db_call("get_approval", approval_id)
            if existing is None:
                raise HTTPException(status_code=404, detail="Approval not found")
            raise HTTPException(
                status_code=409,
                detail=f"Approval is already '{existing.get('status')}' and cannot be resolved",
            )
        return result

    @router.delete("/approvals/{approval_id}", status_code=204)
    async def delete_approval(
        approval_id: str,
        _: bool = Depends(auth_dependency),
    ) -> None:
        existing = await _db_call("get_approval", approval_id)
        if existing is None:
            raise HTTPException(status_code=404, detail="Approval not found")
        deleted = await _db_call("delete_approval", approval_id)
        if not deleted:
            raise HTTPException(status_code=500, detail="Failed to delete approval")

    return router



"""Tool resolution, formatting, and execution helpers for Agent."""

from __future__ import annotations

from collections import deque
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Union,
    cast,
)

if TYPE_CHECKING:
    from agno.agent.agent import Agent

from agno.models.base import Model
from agno.models.message import Message
from agno.models.metrics import MessageMetrics
from agno.models.response import ModelResponse, ModelResponseEvent, ToolExecution
from agno.run import RunContext
from agno.run.agent import RunOutput, RunOutputEvent
from agno.run.messages import RunMessages
from agno.session import AgentSession
from agno.tools import Toolkit
from agno.tools.function import Function
from agno.utils.agent import (
    collect_joint_audios,
    collect_joint_files,
    collect_joint_images,
    collect_joint_videos,
)
from agno.utils.events import (
    create_tool_call_completed_event,
    create_tool_call_error_event,
    create_tool_call_started_event,
    handle_event,
)
from agno.utils.log import log_debug, log_warning


def raise_if_async_tools(agent: Agent) -> None:
    """Raise an exception if any tools contain async functions."""
    if agent.tools is None:
        return

    # Skip check if tools is a callable factory (not yet resolved)
    if not isinstance(agent.tools, list):
        return

    from inspect import iscoroutinefunction

    for tool in agent.tools:
        if isinstance(tool, Toolkit):
            for func in tool.functions:
                if iscoroutinefunction(tool.functions[func].entrypoint):
                    raise Exception(
                        f"Async tool {tool.name} can't be used with synchronous agent.run() or agent.print_response(). "
                        "Use agent.arun() or agent.aprint_response() instead to use this tool."
                    )
        elif isinstance(tool, Function):
            if iscoroutinefunction(tool.entrypoint):
                raise Exception(
                    f"Async function {tool.name} can't be used with synchronous agent.run() or agent.print_response(). "
                    "Use agent.arun() or agent.aprint_response() instead to use this tool."
                )
        elif callable(tool):
            if iscoroutinefunction(tool):
                raise Exception(
                    f"Async function {tool.__name__} can't be used with synchronous agent.run() or agent.print_response(). "
                    "Use agent.arun() or agent.aprint_response() instead to use this tool."
                )


def _raise_if_async_tools_in_list(tools: list) -> None:
    """Raise if any tools in a concrete list are async."""
    from inspect import iscoroutinefunction

    for tool in tools:
        if isinstance(tool, Toolkit):
            for func in tool.functions:
                if iscoroutinefunction(tool.functions[func].entrypoint):
                    raise Exception(
                        f"Async tool {tool.name} can't be used with synchronous agent.run() or agent.print_response(). "
                        "Use agent.arun() or agent.aprint_response() instead to use this tool."
                    )
        elif isinstance(tool, Function):
            if iscoroutinefunction(tool.entrypoint):
                raise Exception(
                    f"Async function {tool.name} can't be used with synchronous agent.run() or agent.print_response(). "
                    "Use agent.arun() or agent.aprint_response() instead to use this tool."
                )
        elif callable(tool):
            if iscoroutinefunction(tool):
                raise Exception(
                    f"Async function {tool.__name__} can't be used with synchronous agent.run() or agent.print_response(). "
                    "Use agent.arun() or agent.aprint_response() instead to use this tool."
                )


def get_tools(
    agent: Agent,
    run_response: RunOutput,
    run_context: RunContext,
    session: AgentSession,
    user_id: Optional[str] = None,
) -> List[Union[Toolkit, Callable, Function, Dict]]:
    from agno.agent import _default_tools, _init
    from agno.utils.callables import (
        get_resolved_knowledge,
        get_resolved_tools,
        resolve_callable_knowledge,
        resolve_callable_tools,
    )

    agent_tools: List[Union[Toolkit, Callable, Function, Dict]] = []

    # Resolve callable factories
    resolve_callable_tools(agent, run_context)
    resolve_callable_knowledge(agent, run_context)

    resolved_tools = get_resolved_tools(agent, run_context)
    resolved_knowledge = get_resolved_knowledge(agent, run_context)

    # Connect tools that require connection management
    _init.connect_connectable_tools(agent)

    # Add provided tools
    if resolved_tools is not None:
        # If not running in async mode, raise if any tool is async
        _raise_if_async_tools_in_list(resolved_tools)
        agent_tools.extend(resolved_tools)

    # Add tools for accessing memory
    if agent.read_chat_history:
        agent_tools.append(_default_tools.get_chat_history_function(agent, session=session))
    if agent.read_tool_call_history:
        agent_tools.append(_default_tools.get_tool_call_history_function(agent, session=session))
    if agent.search_session_history:
        agent_tools.append(
            _default_tools.get_previous_sessions_messages_function(
                agent, num_history_sessions=agent.num_history_sessions, user_id=user_id
            )
        )

    if agent.enable_agentic_memory:
        agent_tools.append(_default_tools.get_update_user_memory_function(agent, user_id=user_id, async_mode=False))

    # Add learning machine tools
    if agent._learning is not None:
        learning_tools = agent._learning.get_tools(
            user_id=user_id,
            session_id=session.session_id if session else None,
            agent_id=agent.id,
        )
        agent_tools.extend(learning_tools)

    if agent.enable_agentic_culture:
        agent_tools.append(_default_tools.get_update_cultural_knowledge_function(agent, async_mode=False))

    if agent.enable_agentic_state:
        agent_tools.append(
            Function(
                name="update_session_state",
                entrypoint=_default_tools.make_update_session_state_entrypoint(agent),
            )
        )

    # Add tools for accessing knowledge
    # Single unified path through get_relevant_docs_from_knowledge(),
    # which checks knowledge_retriever first, then falls back to knowledge.search().
    if (resolved_knowledge is not None or agent.knowledge_retriever is not None) and agent.search_knowledge:
        agent_tools.append(
            _default_tools.create_knowledge_search_tool(
                agent,
                run_response=run_response,
                run_context=run_context,
                knowledge_filters=run_context.knowledge_filters,
                enable_agentic_filters=agent.enable_agentic_knowledge_filters,
                async_mode=False,
            )
        )

    if resolved_knowledge is not None and agent.update_knowledge:
        agent_tools.append(agent.add_to_knowledge)

    # Add tools for accessing skills
    if agent.skills is not None:
        agent_tools.extend(agent.skills.get_tools())

    return agent_tools


async def aget_tools(
    agent: Agent,
    run_response: RunOutput,
    run_context: RunContext,
    session: AgentSession,
    user_id: Optional[str] = None,
    check_mcp_tools: bool = True,
) -> List[Union[Toolkit, Callable, Function, Dict]]:
    from agno.agent import _default_tools, _init
    from agno.utils.callables import (
        aresolve_callable_knowledge,
        aresolve_callable_tools,
        get_resolved_knowledge,
        get_resolved_tools,
    )

    agent_tools: List[Union[Toolkit, Callable, Function, Dict]] = []

    # Resolve callable factories
    await aresolve_callable_tools(agent, run_context)
    await aresolve_callable_knowledge(agent, run_context)

    resolved_tools = get_resolved_tools(agent, run_context)
    resolved_knowledge = get_resolved_knowledge(agent, run_context)

    # Connect tools that require connection management
    _init.connect_connectable_tools(agent)

    # Connect MCP tools
    await _init.connect_mcp_tools(agent)

    # Add provided tools
    if resolved_tools is not None:
        for tool in resolved_tools:
            # Alternate method of using isinstance(tool, (MCPTools, MultiMCPTools)) to avoid imports
            is_mcp_tool = hasattr(type(tool), "__mro__") and any(
                c.__name__ in ["MCPTools", "MultiMCPTools"] for c in type(tool).__mro__
            )

            if is_mcp_tool:
                if tool.refresh_connection:  # type: ignore
                    try:
                        is_alive = await tool.is_alive()  # type: ignore
                        if not is_alive:
                            await tool.connect(force=True)  # type: ignore
                    except (RuntimeError, BaseException) as e:
                        log_warning(f"Failed to check if MCP tool is alive or to connect to it: {e}")
                        continue

                    try:
                        await tool.build_tools()  # type: ignore
                    except (RuntimeError, BaseException) as e:
                        log_warning(f"Failed to build tools for {str(tool)}: {e}")
                        continue

                # Only add the tool if it successfully connected and built its tools
                if check_mcp_tools and not tool.initialized:  # type: ignore
                    continue

            # Add the tool (MCP tools that passed checks, or any non-MCP tool)
            agent_tools.append(tool)

    # Add tools for accessing memory
    if agent.read_chat_history:
        agent_tools.append(_default_tools.get_chat_history_function(agent, session=session))
    if agent.read_tool_call_history:
        agent_tools.append(_default_tools.get_tool_call_history_function(agent, session=session))
    if agent.search_session_history:
        agent_tools.append(
            await _default_tools.aget_previous_sessions_messages_function(
                agent, num_history_sessions=agent.num_history_sessions, user_id=user_id
            )
        )

    if agent.enable_agentic_memory:
        agent_tools.append(_default_tools.get_update_user_memory_function(agent, user_id=user_id, async_mode=True))

    # Add learning machine tools (async)
    if agent._learning is not None:
        learning_tools = await agent._learning.aget_tools(
            user_id=user_id,
            session_id=session.session_id if session else None,
            agent_id=agent.id,
        )
        agent_tools.extend(learning_tools)

    if agent.enable_agentic_culture:
        agent_tools.append(_default_tools.get_update_cultural_knowledge_function(agent, async_mode=True))

    if agent.enable_agentic_state:
        agent_tools.append(
            Function(
                name="update_session_state",
                entrypoint=_default_tools.make_update_session_state_entrypoint(agent),
            )
        )

    # Add tools for accessing knowledge
    # Single unified path through aget_relevant_docs_from_knowledge(),
    # which checks knowledge_retriever first, then falls back to knowledge.search().
    if (resolved_knowledge is not None or agent.knowledge_retriever is not None) and agent.search_knowledge:
        agent_tools.append(
            _default_tools.create_knowledge_search_tool(
                agent,
                run_response=run_response,
                run_context=run_context,
                knowledge_filters=run_context.knowledge_filters,
                enable_agentic_filters=agent.enable_agentic_knowledge_filters,
                async_mode=True,
            )
        )

    if resolved_knowledge is not None and agent.update_knowledge:
        agent_tools.append(agent.add_to_knowledge)

    # Add tools for accessing skills
    if agent.skills is not None:
        agent_tools.extend(agent.skills.get_tools())

    return agent_tools


def parse_tools(
    agent: Agent,
    tools: List[Union[Toolkit, Callable, Function, Dict]],
    model: Model,
    run_context: Optional[RunContext] = None,
    async_mode: bool = False,
) -> List[Union[Function, dict]]:
    _function_names: List[str] = []
    _functions: List[Union[Function, dict]] = []
    agent._tool_instructions = []

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    # Check if we need strict mode for the functions for the model
    strict = False
    if (
        output_schema is not None
        and (agent.structured_outputs or (not agent.use_json_mode))
        and model.supports_native_structured_outputs
    ):
        strict = True

    for tool in tools:
        if isinstance(tool, Dict):
            # If a dict is passed, it is a builtin tool
            # that is run by the model provider and not the Agent
            _functions.append(tool)
            log_debug(f"Included builtin tool {tool}")

        elif isinstance(tool, Toolkit):
            # For each function in the toolkit and process entrypoint
            toolkit_functions = tool.get_async_functions() if async_mode else tool.get_functions()
            for name, _func in toolkit_functions.items():
                if name in _function_names:
                    continue
                _function_names.append(name)
                _func = _func.model_copy(deep=True)
                _func._agent = agent
                # Respect the function's explicit strict setting if set
                effective_strict = strict if _func.strict is None else _func.strict
                _func.process_entrypoint(strict=effective_strict)
                if strict and _func.strict is None:
                    _func.strict = True
                if agent.tool_hooks is not None:
                    _func.tool_hooks = agent.tool_hooks
                _functions.append(_func)
                log_debug(f"Added tool {name} from {tool.name}")

            # Add instructions from the toolkit
            if tool.add_instructions and tool.instructions is not None:
                agent._tool_instructions.append(tool.instructions)

        elif isinstance(tool, Function):
            if tool.name in _function_names:
                continue
            _function_names.append(tool.name)

            tool = tool.model_copy(deep=True)
            # Respect the function's explicit strict setting if set
            effective_strict = strict if tool.strict is None else tool.strict
            tool.process_entrypoint(strict=effective_strict)

            tool._agent = agent
            if strict and tool.strict is None:
                tool.strict = True
            if agent.tool_hooks is not None:
                tool.tool_hooks = agent.tool_hooks
            _functions.append(tool)
            log_debug(f"Added tool {tool.name}")

            # Add instructions from the Function
            if tool.add_instructions and tool.instructions is not None:
                agent._tool_instructions.append(tool.instructions)

        elif callable(tool):
            try:
                function_name = tool.__name__

                if function_name in _function_names:
                    continue
                _function_names.append(function_name)

                _func = Function.from_callable(tool, strict=strict)
                # Detect @approval sentinel on raw callable
                _approval_type = getattr(tool, "_agno_approval_type", None)
                if _approval_type is not None:
                    _func.approval_type = _approval_type
                    if _approval_type == "required" and not any(
                        [_func.requires_user_input, _func.requires_confirmation, _func.external_execution]
                    ):
                        _func.requires_confirmation = True
                    elif _approval_type == "audit" and not any(
                        [_func.requires_user_input, _func.requires_confirmation, _func.external_execution]
                    ):
                        raise ValueError(
                            "@approval(type='audit') requires at least one HITL flag "
                            "('requires_confirmation', 'requires_user_input', or 'external_execution') "
                            "to be set on @tool()."
                        )
                _func = _func.model_copy(deep=True)
                _func._agent = agent
                if strict:
                    _func.strict = True
                if agent.tool_hooks is not None:
                    _func.tool_hooks = agent.tool_hooks
                _functions.append(_func)
                log_debug(f"Added tool {_func.name}")
            except Exception as e:
                log_warning(f"Could not add tool {tool}: {e}")

    return _functions


def determine_tools_for_model(
    agent: Agent,
    model: Model,
    processed_tools: List[Union[Toolkit, Callable, Function, Dict]],
    run_response: RunOutput,
    run_context: RunContext,
    session: AgentSession,
    async_mode: bool = False,
) -> List[Union[Function, dict]]:
    _functions: List[Union[Function, dict]] = []

    # Get Agent tools
    if processed_tools is not None and len(processed_tools) > 0:
        log_debug("Processing tools for model")
        _functions = parse_tools(
            agent, tools=processed_tools, model=model, run_context=run_context, async_mode=async_mode
        )

    # Update the session state for the functions
    if _functions:
        from inspect import signature

        # Check if any functions need media before collecting
        needs_media = any(
            any(param in signature(func.entrypoint).parameters for param in ["images", "videos", "audios", "files"])
            for func in _functions
            if isinstance(func, Function) and func.entrypoint is not None
        )

        # Only collect media if functions actually need them
        joint_images = collect_joint_images(run_response.input, session) if needs_media else None
        joint_files = collect_joint_files(run_response.input) if needs_media else None
        joint_audios = collect_joint_audios(run_response.input, session) if needs_media else None
        joint_videos = collect_joint_videos(run_response.input, session) if needs_media else None

        for func in _functions:  # type: ignore
            if isinstance(func, Function):
                func._run_context = run_context
                func._images = joint_images
                func._files = joint_files
                func._audios = joint_audios
                func._videos = joint_videos

    return _functions


# ---------------------------------------------------------------------------
# Tool Execution
# ---------------------------------------------------------------------------


def handle_external_execution_update(agent: Agent, run_messages: RunMessages, tool: ToolExecution):
    agent.model = cast(Model, agent.model)

    if tool.result is not None:
        for msg in run_messages.messages:
            # Skip if the message is already in the run_messages
            if msg.tool_call_id == tool.tool_call_id:
                break
        else:
            run_messages.messages.append(
                Message(
                    role=agent.model.tool_message_role,
                    content=tool.result,
                    tool_call_id=tool.tool_call_id,
                    tool_name=tool.tool_name,
                    tool_args=tool.tool_args,
                    tool_call_error=tool.tool_call_error,
                    stop_after_tool_call=tool.stop_after_tool_call,
                )
            )
        tool.external_execution_required = False
    else:
        raise ValueError(f"Tool {tool.tool_name} requires external execution, cannot continue run")


def handle_user_input_update(agent: Agent, tool: ToolExecution):
    for field in tool.user_input_schema or []:
        if not tool.tool_args:
            tool.tool_args = {}
        tool.tool_args[field.name] = field.value


def handle_get_user_input_tool_update(agent: Agent, run_messages: RunMessages, tool: ToolExecution):
    import json

    agent.model = cast(Model, agent.model)
    # Skipping tool without user_input_schema so that tool_call_id is not repeated
    if not hasattr(tool, "user_input_schema") or not tool.user_input_schema:
        return
    user_input_result = [
        {"name": user_input_field.name, "value": user_input_field.value}
        for user_input_field in tool.user_input_schema or []
    ]
    # Add the tool call result to the run_messages
    run_messages.messages.append(
        Message(
            role=agent.model.tool_message_role,
            content=f"User inputs retrieved: {json.dumps(user_input_result)}",
            tool_call_id=tool.tool_call_id,
            tool_name=tool.tool_name,
            tool_args=tool.tool_args,
            metrics=MessageMetrics(duration=0),
        )
    )


def handle_ask_user_tool_update(agent: Agent, run_messages: RunMessages, tool: ToolExecution):
    import json

    agent.model = cast(Model, agent.model)
    if not hasattr(tool, "user_feedback_schema") or not tool.user_feedback_schema:
        return
    feedback_result = [
        {"question": q.question, "selected": q.selected_options or []} for q in tool.user_feedback_schema
    ]
    run_messages.messages.append(
        Message(
            role=agent.model.tool_message_role,
            content=f"User feedback received: {json.dumps(feedback_result)}",
            tool_call_id=tool.tool_call_id,
            tool_name=tool.tool_name,
            tool_args=tool.tool_args,
            metrics=MessageMetrics(duration=0),
        )
    )


def _maybe_create_audit_approval(
    agent: "Agent", tool_execution: ToolExecution, run_response: RunOutput, status: str
) -> None:
    """Create an audit approval record if the tool has approval_type='audit'."""
    if getattr(tool_execution, "approval_type", None) == "audit":
        from agno.run.approval import create_audit_approval

        create_audit_approval(
            db=agent.db,
            tool_execution=tool_execution,
            run_response=run_response,
            status=status,
            agent_id=agent.id,
            agent_name=agent.name,
        )


async def _amaybe_create_audit_approval(
    agent: "Agent", tool_execution: ToolExecution, run_response: RunOutput, status: str
) -> None:
    """Async: create an audit approval record if the tool has approval_type='audit'."""
    if getattr(tool_execution, "approval_type", None) == "audit":
        from agno.run.approval import acreate_audit_approval

        await acreate_audit_approval(
            db=agent.db,
            tool_execution=tool_execution,
            run_response=run_response,
            status=status,
            agent_id=agent.id,
            agent_name=agent.name,
        )


def run_tool(
    agent: Agent,
    run_response: RunOutput,
    run_messages: RunMessages,
    tool: ToolExecution,
    functions: Optional[Dict[str, Function]] = None,
    stream_events: bool = False,
) -> Iterator[RunOutputEvent]:
    from agno.run.agent import CustomEvent

    agent.model = cast(Model, agent.model)
    # Execute the tool
    function_call = agent.model.get_function_call_to_run_from_tool_execution(tool, functions)
    function_call_results: List[Message] = []

    for call_result in agent.model.run_function_call(
        function_call=function_call,
        function_call_results=function_call_results,
    ):
        if isinstance(call_result, ModelResponse):
            if call_result.event == ModelResponseEvent.tool_call_started.value:
                if stream_events:
                    yield handle_event(  # type: ignore
                        create_tool_call_started_event(from_run_response=run_response, tool=tool),
                        run_response,
                        events_to_skip=agent.events_to_skip,  # type: ignore
                        store_events=agent.store_events,
                    )

            if call_result.event == ModelResponseEvent.tool_call_completed.value and call_result.tool_executions:
                tool_execution = call_result.tool_executions[0]
                tool.result = tool_execution.result
                tool.tool_call_error = tool_execution.tool_call_error
                if stream_events:
                    yield handle_event(  # type: ignore
                        create_tool_call_completed_event(
                            from_run_response=run_response, tool=tool, content=call_result.content
                        ),
                        run_response,
                        events_to_skip=agent.events_to_skip,  # type: ignore
                        store_events=agent.store_events,
                    )
                    if tool.tool_call_error:
                        yield handle_event(  # type: ignore
                            create_tool_call_error_event(
                                from_run_response=run_response, tool=tool, error=str(tool.result)
                            ),
                            run_response,
                            events_to_skip=agent.events_to_skip,  # type: ignore
                            store_events=agent.store_events,
                        )
        # Yield CustomEvent instances from sync tool generators
        elif isinstance(call_result, CustomEvent):
            if stream_events:
                yield call_result  # type: ignore

    if len(function_call_results) > 0:
        run_messages.messages.extend(function_call_results)


def reject_tool_call(
    agent: Agent, run_messages: RunMessages, tool: ToolExecution, functions: Optional[Dict[str, Function]] = None
):
    agent.model = cast(Model, agent.model)
    function_call = agent.model.get_function_call_to_run_from_tool_execution(tool, functions)
    function_call.error = tool.confirmation_note or "Function call was rejected by the user"
    function_call_result = agent.model.create_function_call_result(
        function_call=function_call,
        success=False,
    )
    run_messages.messages.append(function_call_result)


async def arun_tool(
    agent: Agent,
    run_response: RunOutput,
    run_messages: RunMessages,
    tool: ToolExecution,
    functions: Optional[Dict[str, Function]] = None,
    stream_events: bool = False,
) -> AsyncIterator[RunOutputEvent]:
    from agno.run.agent import CustomEvent

    agent.model = cast(Model, agent.model)

    # Execute the tool
    function_call = agent.model.get_function_call_to_run_from_tool_execution(tool, functions)
    function_call_results: List[Message] = []

    async for call_result in agent.model.arun_function_calls(
        function_calls=[function_call],
        function_call_results=function_call_results,
        skip_pause_check=True,
    ):
        if isinstance(call_result, ModelResponse):
            if call_result.event == ModelResponseEvent.tool_call_started.value:
                if stream_events:
                    yield handle_event(  # type: ignore
                        create_tool_call_started_event(from_run_response=run_response, tool=tool),
                        run_response,
                        events_to_skip=agent.events_to_skip,  # type: ignore
                        store_events=agent.store_events,
                    )
            if call_result.event == ModelResponseEvent.tool_call_completed.value and call_result.tool_executions:
                tool_execution = call_result.tool_executions[0]
                tool.result = tool_execution.result
                tool.tool_call_error = tool_execution.tool_call_error
                if stream_events:
                    yield handle_event(  # type: ignore
                        create_tool_call_completed_event(
                            from_run_response=run_response, tool=tool, content=call_result.content
                        ),
                        run_response,
                        events_to_skip=agent.events_to_skip,  # type: ignore
                        store_events=agent.store_events,
                    )
                    if tool.tool_call_error:
                        yield handle_event(  # type: ignore
                            create_tool_call_error_event(
                                from_run_response=run_response, tool=tool, error=str(tool.result)
                            ),
                            run_response,
                            events_to_skip=agent.events_to_skip,  # type: ignore
                            store_events=agent.store_events,
                        )
        # Yield CustomEvent instances from async tool generators
        elif isinstance(call_result, CustomEvent):
            if stream_events:
                yield call_result  # type: ignore

    if len(function_call_results) > 0:
        run_messages.messages.extend(function_call_results)


def handle_tool_call_updates(
    agent: Agent, run_response: RunOutput, run_messages: RunMessages, tools: List[Union[Function, dict]]
):
    agent.model = cast(Model, agent.model)
    _functions = {tool.name: tool for tool in tools if isinstance(tool, Function)}

    for _t in run_response.tools or []:
        # Case 1: Handle confirmed tools and execute them
        if _t.requires_confirmation is not None and _t.requires_confirmation is True and _functions:
            # Tool is confirmed and hasn't been run before
            if _t.confirmed is not None and _t.confirmed is True and _t.result is None:
                # Consume the generator without yielding
                deque(run_tool(agent, run_response, run_messages, _t, functions=_functions), maxlen=0)
            else:
                reject_tool_call(agent, run_messages, _t, functions=_functions)
                _t.confirmed = False
                _t.confirmation_note = _t.confirmation_note or "Tool call was rejected"
                _t.tool_call_error = True
            _maybe_create_audit_approval(agent, _t, run_response, "approved" if _t.confirmed is True else "rejected")
            _t.requires_confirmation = False

        # Case 2: Handle external execution required tools
        elif _t.external_execution_required is not None and _t.external_execution_required is True:
            handle_external_execution_update(agent, run_messages=run_messages, tool=_t)
            _maybe_create_audit_approval(agent, _t, run_response, "approved")

        # Case 3a: Agentic user input required
        elif _t.tool_name == "get_user_input" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_get_user_input_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True

        # Case 3b: User feedback (ask_user) required
        elif _t.tool_name == "ask_user" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_ask_user_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True

        # Case 4: Handle user input required tools
        elif _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_user_input_update(agent, tool=_t)
            _t.requires_user_input = False
            _t.answered = True
            # Consume the generator without yielding
            deque(run_tool(agent, run_response, run_messages, _t, functions=_functions), maxlen=0)
            _maybe_create_audit_approval(agent, _t, run_response, "approved")


def handle_tool_call_updates_stream(
    agent: Agent,
    run_response: RunOutput,
    run_messages: RunMessages,
    tools: List[Union[Function, dict]],
    stream_events: bool = False,
) -> Iterator[RunOutputEvent]:
    agent.model = cast(Model, agent.model)
    _functions = {tool.name: tool for tool in tools if isinstance(tool, Function)}

    for _t in run_response.tools or []:
        # Case 1: Handle confirmed tools and execute them
        if _t.requires_confirmation is not None and _t.requires_confirmation is True and _functions:
            # Tool is confirmed and hasn't been run before
            if _t.confirmed is not None and _t.confirmed is True and _t.result is None:
                yield from run_tool(
                    agent, run_response, run_messages, _t, functions=_functions, stream_events=stream_events
                )
            else:
                reject_tool_call(agent, run_messages, _t, functions=_functions)
                _t.confirmed = False
                _t.confirmation_note = _t.confirmation_note or "Tool call was rejected"
                _t.tool_call_error = True
            _maybe_create_audit_approval(agent, _t, run_response, "approved" if _t.confirmed is True else "rejected")
            _t.requires_confirmation = False

        # Case 2: Handle external execution required tools
        elif _t.external_execution_required is not None and _t.external_execution_required is True:
            handle_external_execution_update(agent, run_messages=run_messages, tool=_t)
            _maybe_create_audit_approval(agent, _t, run_response, "approved")

        # Case 3a: Agentic user input required
        elif _t.tool_name == "get_user_input" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_get_user_input_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True

        # Case 3b: User feedback (ask_user) required
        elif _t.tool_name == "ask_user" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_ask_user_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True

        # Case 4: Handle user input required tools
        elif _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_user_input_update(agent, tool=_t)
            yield from run_tool(
                agent, run_response, run_messages, _t, functions=_functions, stream_events=stream_events
            )
            _t.requires_user_input = False
            _t.answered = True
            _maybe_create_audit_approval(agent, _t, run_response, "approved")


async def ahandle_tool_call_updates(
    agent: Agent, run_response: RunOutput, run_messages: RunMessages, tools: List[Union[Function, dict]]
):
    agent.model = cast(Model, agent.model)
    _functions = {tool.name: tool for tool in tools if isinstance(tool, Function)}

    for _t in run_response.tools or []:
        # Case 1: Handle confirmed tools and execute them
        if _t.requires_confirmation is not None and _t.requires_confirmation is True and _functions:
            # Tool is confirmed and hasn't been run before
            if _t.confirmed is not None and _t.confirmed is True and _t.result is None:
                async for _ in arun_tool(agent, run_response, run_messages, _t, functions=_functions):
                    pass
            else:
                reject_tool_call(agent, run_messages, _t, functions=_functions)
                _t.confirmed = False
                _t.confirmation_note = _t.confirmation_note or "Tool call was rejected"
                _t.tool_call_error = True
            await _amaybe_create_audit_approval(
                agent, _t, run_response, "approved" if _t.confirmed is True else "rejected"
            )
            _t.requires_confirmation = False

        # Case 2: Handle external execution required tools
        elif _t.external_execution_required is not None and _t.external_execution_required is True:
            handle_external_execution_update(agent, run_messages=run_messages, tool=_t)
            await _amaybe_create_audit_approval(agent, _t, run_response, "approved")
        # Case 3a: Agentic user input required
        elif _t.tool_name == "get_user_input" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_get_user_input_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True
        # Case 3b: User feedback (ask_user) required
        elif _t.tool_name == "ask_user" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_ask_user_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True
        # Case 4: Handle user input required tools
        elif _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_user_input_update(agent, tool=_t)
            async for _ in arun_tool(agent, run_response, run_messages, _t, functions=_functions):
                pass
            _t.requires_user_input = False
            _t.answered = True
            await _amaybe_create_audit_approval(agent, _t, run_response, "approved")


async def ahandle_tool_call_updates_stream(
    agent: Agent,
    run_response: RunOutput,
    run_messages: RunMessages,
    tools: List[Union[Function, dict]],
    stream_events: bool = False,
) -> AsyncIterator[RunOutputEvent]:
    agent.model = cast(Model, agent.model)
    _functions = {tool.name: tool for tool in tools if isinstance(tool, Function)}

    for _t in run_response.tools or []:
        # Case 1: Handle confirmed tools and execute them
        if _t.requires_confirmation is not None and _t.requires_confirmation is True and _functions:
            # Tool is confirmed and hasn't been run before
            if _t.confirmed is not None and _t.confirmed is True and _t.result is None:
                async for event in arun_tool(
                    agent, run_response, run_messages, _t, functions=_functions, stream_events=stream_events
                ):
                    yield event
            else:
                reject_tool_call(agent, run_messages, _t, functions=_functions)
                _t.confirmed = False
                _t.confirmation_note = _t.confirmation_note or "Tool call was rejected"
                _t.tool_call_error = True
            await _amaybe_create_audit_approval(
                agent, _t, run_response, "approved" if _t.confirmed is True else "rejected"
            )
            _t.requires_confirmation = False

        # Case 2: Handle external execution required tools
        elif _t.external_execution_required is not None and _t.external_execution_required is True:
            handle_external_execution_update(agent, run_messages=run_messages, tool=_t)
            await _amaybe_create_audit_approval(agent, _t, run_response, "approved")
        # Case 3a: Agentic user input required
        elif _t.tool_name == "get_user_input" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_get_user_input_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True
        # Case 3b: User feedback (ask_user) required
        elif _t.tool_name == "ask_user" and _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_ask_user_tool_update(agent, run_messages=run_messages, tool=_t)
            _t.requires_user_input = False
            _t.answered = True
        # Case 4: Handle user input required tools
        elif _t.requires_user_input is not None and _t.requires_user_input is True:
            handle_user_input_update(agent, tool=_t)
            async for event in arun_tool(
                agent, run_response, run_messages, _t, functions=_functions, stream_events=stream_events
            ):
                yield event
            _t.requires_user_input = False
            _t.answered = True
            await _amaybe_create_audit_approval(agent, _t, run_response, "approved")



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Type
from uuid import uuid4

from pydantic import BaseModel

from agno.db.base import BaseDb
from agno.models.base import Model
from agno.tools.function import Function
from agno.tools.toolkit import Toolkit
from agno.vectordb.base import VectorDb

if TYPE_CHECKING:
    from agno.agent import Agent
    from agno.team import Team


@dataclass
class Registry:
    """
    Registry is used to manage non serializable objects like tools, models, databases, vector databases,
    agents, and teams.
    """

    name: Optional[str] = None
    description: Optional[str] = None
    id: str = field(default_factory=lambda: str(uuid4()))
    tools: List[Any] = field(default_factory=list)
    models: List[Model] = field(default_factory=list)
    dbs: List[BaseDb] = field(default_factory=list)
    vector_dbs: List[VectorDb] = field(default_factory=list)
    schemas: List[Type[BaseModel]] = field(default_factory=list)
    functions: List[Callable] = field(default_factory=list)
    # Code-defined agents and teams (for workflow rehydration)
    agents: List[Agent] = field(default_factory=list)
    teams: List[Team] = field(default_factory=list)

    @cached_property
    def _entrypoint_lookup(self) -> Dict[str, Callable]:
        lookup: Dict[str, Callable] = {}
        for tool in self.tools:
            if isinstance(tool, Toolkit):
                for func in tool.functions.values():
                    if func.entrypoint is not None:
                        lookup[func.name] = func.entrypoint
            elif isinstance(tool, Function):
                if tool.entrypoint is not None:
                    lookup[tool.name] = tool.entrypoint
            elif callable(tool):
                lookup[tool.__name__] = tool
        return lookup

    def rehydrate_function(self, func_dict: Dict[str, Any]) -> Function:
        """Reconstruct a Function from dict, reattaching its entrypoint."""
        func = Function.from_dict(func_dict)
        func.entrypoint = self._entrypoint_lookup.get(func.name)
        return func

    def get_schema(self, name: str) -> Optional[Type[BaseModel]]:
        """Get a schema by name."""
        if self.schemas:
            return next((s for s in self.schemas if s.__name__ == name), None)
        return None

    def get_db(self, db_id: str) -> Optional[BaseDb]:
        """Get a database by id from the registry.

        Args:
            db_id: The database id to look up

        Returns:
            The database instance if found, None otherwise
        """
        if self.dbs:
            return next((db for db in self.dbs if db.id == db_id), None)
        return None

    def get_function(self, name: str) -> Optional[Callable]:
        return next((f for f in self.functions if f.__name__ == name), None)

    def get_agent(self, agent_id: str) -> Optional[Agent]:
        """Get an agent by id from the registry."""
        if self.agents:
            return next((a for a in self.agents if getattr(a, "id", None) == agent_id), None)
        return None

    def get_team(self, team_id: str) -> Optional[Team]:
        """Get a team by id from the registry."""
        if self.teams:
            return next((t for t in self.teams if getattr(t, "id", None) == team_id), None)
        return None

    def get_agent_ids(self) -> Set[str]:
        """Get the set of all agent IDs in this registry."""
        if self.agents:
            return {aid for a in self.agents if (aid := getattr(a, "id", None)) is not None}
        return set()

    def get_team_ids(self) -> Set[str]:
        """Get the set of all team IDs in this registry."""
        if self.teams:
            return {tid for t in self.teams if (tid := getattr(t, "id", None)) is not None}
        return set()

    def get_all_component_ids(self) -> Set[str]:
        """Get the set of all agent and team IDs in this registry."""
        return self.get_agent_ids() | self.get_team_ids()



import logging
import math
from typing import List, Optional, Union, cast
from uuid import uuid4

from fastapi import Depends, HTTPException, Path, Query, Request
from fastapi.routing import APIRouter

from agno.db.base import AsyncBaseDb, BaseDb
from agno.db.schemas import UserMemory
from agno.models.utils import get_model
from agno.os.auth import get_auth_token_from_request, get_authentication_dependency
from agno.os.routers.memory.schemas import (
    DeleteMemoriesRequest,
    OptimizeMemoriesRequest,
    OptimizeMemoriesResponse,
    UserMemoryCreateSchema,
    UserMemorySchema,
    UserStatsSchema,
)
from agno.os.schema import (
    BadRequestResponse,
    InternalServerErrorResponse,
    NotFoundResponse,
    PaginatedResponse,
    PaginationInfo,
    SortOrder,
    UnauthenticatedResponse,
    ValidationErrorResponse,
)
from agno.os.settings import AgnoAPISettings
from agno.os.utils import get_db
from agno.remote.base import RemoteDb

logger = logging.getLogger(__name__)


def get_memory_router(
    dbs: dict[str, list[Union[BaseDb, AsyncBaseDb, RemoteDb]]], settings: AgnoAPISettings = AgnoAPISettings(), **kwargs
) -> APIRouter:
    """Create memory router with comprehensive OpenAPI documentation for user memory management endpoints."""
    router = APIRouter(
        dependencies=[Depends(get_authentication_dependency(settings))],
        tags=["Memory"],
        responses={
            400: {"description": "Bad Request", "model": BadRequestResponse},
            401: {"description": "Unauthorized", "model": UnauthenticatedResponse},
            404: {"description": "Not Found", "model": NotFoundResponse},
            422: {"description": "Validation Error", "model": ValidationErrorResponse},
            500: {"description": "Internal Server Error", "model": InternalServerErrorResponse},
        },
    )
    return attach_routes(router=router, dbs=dbs)


def attach_routes(router: APIRouter, dbs: dict[str, list[Union[BaseDb, AsyncBaseDb, RemoteDb]]]) -> APIRouter:
    @router.post(
        "/memories",
        response_model=UserMemorySchema,
        status_code=200,
        operation_id="create_memory",
        summary="Create Memory",
        description=(
            "Create a new user memory with content and associated topics. "
            "Memories are used to store contextual information for users across conversations."
        ),
        responses={
            200: {
                "description": "Memory created successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "memory_id": "mem-123",
                            "memory": "User prefers technical explanations with code examples",
                            "topics": ["preferences", "communication_style", "technical"],
                            "user_id": "user-456",
                            "created_at": "2024-01-15T10:30:00Z",
                            "updated_at": "2024-01-15T10:30:00Z",
                        }
                    }
                },
            },
            400: {"description": "Invalid request data", "model": BadRequestResponse},
            422: {"description": "Validation error in payload", "model": ValidationErrorResponse},
            500: {"description": "Failed to create memory", "model": InternalServerErrorResponse},
        },
    )
    async def create_memory(
        request: Request,
        payload: UserMemoryCreateSchema,
        db_id: Optional[str] = Query(default=None, description="Database ID to use for memory storage"),
        table: Optional[str] = Query(default=None, description="Table to use for memory storage"),
    ) -> UserMemorySchema:
        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id
            payload.user_id = user_id

        if payload.user_id is None:
            raise HTTPException(status_code=400, detail="User ID is required")

        db = await get_db(dbs, db_id, table)

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.create_memory(
                memory=payload.memory,
                topics=payload.topics or [],
                user_id=payload.user_id,
                db_id=db_id,
                table=table,
                headers=headers,
            )

        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            user_memory = await db.upsert_user_memory(
                memory=UserMemory(
                    memory_id=str(uuid4()),
                    memory=payload.memory,
                    topics=payload.topics or [],
                    user_id=payload.user_id,
                ),
                deserialize=False,
            )
        else:
            user_memory = db.upsert_user_memory(
                memory=UserMemory(
                    memory_id=str(uuid4()),
                    memory=payload.memory,
                    topics=payload.topics or [],
                    user_id=payload.user_id,
                ),
                deserialize=False,
            )

        if not user_memory:
            raise HTTPException(status_code=500, detail="Failed to create memory")

        return UserMemorySchema.from_dict(user_memory)  # type: ignore

    @router.delete(
        "/memories/{memory_id}",
        status_code=204,
        operation_id="delete_memory",
        summary="Delete Memory",
        description="Permanently delete a specific user memory. This action cannot be undone.",
        responses={
            204: {"description": "Memory deleted successfully"},
            404: {"description": "Memory not found", "model": NotFoundResponse},
            500: {"description": "Failed to delete memory", "model": InternalServerErrorResponse},
        },
    )
    async def delete_memory(
        request: Request,
        memory_id: str = Path(description="Memory ID to delete"),
        user_id: Optional[str] = Query(default=None, description="User ID to delete memory for"),
        db_id: Optional[str] = Query(default=None, description="Database ID to use for deletion"),
        table: Optional[str] = Query(default=None, description="Table to use for deletion"),
    ) -> None:
        db = await get_db(dbs, db_id, table)

        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.delete_memory(
                memory_id=memory_id,
                user_id=user_id,
                db_id=db_id,
                table=table,
                headers=headers,
            )
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            await db.delete_user_memory(memory_id=memory_id, user_id=user_id)
        else:
            db.delete_user_memory(memory_id=memory_id, user_id=user_id)

    @router.delete(
        "/memories",
        status_code=204,
        operation_id="delete_memories",
        summary="Delete Multiple Memories",
        description=(
            "Delete multiple user memories by their IDs in a single operation. "
            "This action cannot be undone and all specified memories will be permanently removed."
        ),
        responses={
            204: {"description": "Memories deleted successfully"},
            400: {"description": "Invalid request - empty memory_ids list", "model": BadRequestResponse},
            500: {"description": "Failed to delete memories", "model": InternalServerErrorResponse},
        },
    )
    async def delete_memories(
        http_request: Request,
        request: DeleteMemoriesRequest,
        db_id: Optional[str] = Query(default=None, description="Database ID to use for deletion"),
        table: Optional[str] = Query(default=None, description="Table to use for deletion"),
    ) -> None:
        db = await get_db(dbs, db_id, table)

        if hasattr(http_request.state, "user_id") and http_request.state.user_id is not None:
            request.user_id = http_request.state.user_id

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(http_request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.delete_memories(
                memory_ids=request.memory_ids,
                user_id=request.user_id,
                db_id=db_id,
                table=table,
                headers=headers,
            )

        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            await db.delete_user_memories(memory_ids=request.memory_ids, user_id=request.user_id)
        else:
            db.delete_user_memories(memory_ids=request.memory_ids, user_id=request.user_id)

    @router.get(
        "/memories",
        response_model=PaginatedResponse[UserMemorySchema],
        status_code=200,
        operation_id="get_memories",
        summary="List Memories",
        description=(
            "Retrieve paginated list of user memories with filtering and search capabilities. "
            "Filter by user, agent, team, topics, or search within memory content."
        ),
        responses={
            200: {
                "description": "Memories retrieved successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "data": [
                                {
                                    "memory_id": "f9361a69-2997-40c7-ae4e-a5861d434047",
                                    "memory": "User likes coffee.",
                                    "topics": ["preferences"],
                                    "agent_id": None,
                                    "team_id": None,
                                    "user_id": "123",
                                    "updated_at": "2025-09-01T07:53:17Z",
                                }
                            ]
                        }
                    }
                },
            }
        },
    )
    async def get_memories(
        request: Request,
        user_id: Optional[str] = Query(default=None, description="Filter memories by user ID"),
        agent_id: Optional[str] = Query(default=None, description="Filter memories by agent ID"),
        team_id: Optional[str] = Query(default=None, description="Filter memories by team ID"),
        topics: Optional[List[str]] = Depends(parse_topics),
        search_content: Optional[str] = Query(default=None, description="Fuzzy search within memory content"),
        limit: Optional[int] = Query(default=20, description="Number of memories to return per page", ge=1),
        page: Optional[int] = Query(default=1, description="Page number for pagination", ge=0),
        sort_by: Optional[str] = Query(default="updated_at", description="Field to sort memories by"),
        sort_order: Optional[SortOrder] = Query(default="desc", description="Sort order (asc or desc)"),
        db_id: Optional[str] = Query(default=None, description="Database ID to query memories from"),
        table: Optional[str] = Query(default=None, description="The database table to use"),
    ) -> PaginatedResponse[UserMemorySchema]:
        db = await get_db(dbs, db_id, table)

        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.get_memories(
                user_id=user_id,
                agent_id=agent_id,
                team_id=team_id,
                topics=topics,
                search_content=search_content,
                limit=limit,
                page=page,
                sort_by=sort_by,
                sort_order=sort_order.value if sort_order else "desc",
                db_id=db_id,
                table=table,
                headers=headers,
            )

        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            user_memories, total_count = await db.get_user_memories(
                limit=limit,
                page=page,
                user_id=user_id,
                agent_id=agent_id,
                team_id=team_id,
                topics=topics,
                search_content=search_content,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )
        else:
            user_memories, total_count = db.get_user_memories(  # type: ignore
                limit=limit,
                page=page,
                user_id=user_id,
                agent_id=agent_id,
                team_id=team_id,
                topics=topics,
                search_content=search_content,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )

        memories = [UserMemorySchema.from_dict(user_memory) for user_memory in user_memories]  # type: ignore
        return PaginatedResponse(
            data=[memory for memory in memories if memory is not None],
            meta=PaginationInfo(
                page=page,
                limit=limit,
                total_count=total_count,  # type: ignore
                total_pages=math.ceil(total_count / limit) if limit is not None and limit > 0 else 0,  # type: ignore
            ),
        )

    @router.get(
        "/memories/{memory_id}",
        response_model=UserMemorySchema,
        status_code=200,
        operation_id="get_memory",
        summary="Get Memory by ID",
        description="Retrieve detailed information about a specific user memory by its ID.",
        responses={
            200: {
                "description": "Memory retrieved successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "memory_id": "f9361a69-2997-40c7-ae4e-a5861d434047",
                            "memory": "User likes coffee.",
                            "topics": ["preferences"],
                            "agent_id": None,
                            "team_id": None,
                            "user_id": "123",
                            "updated_at": "2025-09-01T07:53:17Z",
                        }
                    }
                },
            },
            404: {"description": "Memory not found", "model": NotFoundResponse},
        },
    )
    async def get_memory(
        request: Request,
        memory_id: str = Path(description="Memory ID to retrieve"),
        user_id: Optional[str] = Query(default=None, description="User ID to query memory for"),
        db_id: Optional[str] = Query(default=None, description="Database ID to query memory from"),
        table: Optional[str] = Query(default=None, description="Table to query memory from"),
    ) -> UserMemorySchema:
        db = await get_db(dbs, db_id, table)

        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.get_memory(
                memory_id=memory_id,
                user_id=user_id,
                db_id=db_id,
                table=table,
                headers=headers,
            )

        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            user_memory = await db.get_user_memory(memory_id=memory_id, user_id=user_id, deserialize=False)
        else:
            user_memory = db.get_user_memory(memory_id=memory_id, user_id=user_id, deserialize=False)
        if not user_memory:
            raise HTTPException(status_code=404, detail=f"Memory with ID {memory_id} not found")

        return UserMemorySchema.from_dict(user_memory)  # type: ignore

    @router.get(
        "/memory_topics",
        response_model=List[str],
        status_code=200,
        operation_id="get_memory_topics",
        summary="Get Memory Topics",
        description=(
            "Retrieve all unique topics associated with memories in the system. "
            "Useful for filtering and categorizing memories by topic."
        ),
        responses={
            200: {
                "description": "Memory topics retrieved successfully",
                "content": {
                    "application/json": {
                        "example": [
                            "preferences",
                            "communication_style",
                            "technical",
                            "industry",
                            "compliance",
                            "code_examples",
                            "requirements",
                            "healthcare",
                            "finance",
                        ]
                    }
                },
            }
        },
    )
    async def get_topics(
        request: Request,
        user_id: Optional[str] = Query(default=None, description="User ID to filter topics for"),
        db_id: Optional[str] = Query(default=None, description="Database ID to query topics from"),
        table: Optional[str] = Query(default=None, description="Table to query topics from"),
    ) -> List[str]:
        db = await get_db(dbs, db_id, table)

        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.get_memory_topics(
                db_id=db_id,
                table=table,
                headers=headers,
            )

        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            return await db.get_all_memory_topics(user_id=user_id)
        else:
            return db.get_all_memory_topics(user_id=user_id)

    @router.patch(
        "/memories/{memory_id}",
        response_model=UserMemorySchema,
        status_code=200,
        operation_id="update_memory",
        summary="Update Memory",
        description=(
            "Update an existing user memory's content and topics. "
            "Replaces the entire memory content and topic list with the provided values."
        ),
        responses={
            200: {
                "description": "Memory updated successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "memory_id": "f9361a69-2997-40c7-ae4e-a5861d434047",
                            "memory": "User likes coffee.",
                            "topics": ["preferences"],
                            "agent_id": None,
                            "team_id": None,
                            "user_id": "123",
                            "updated_at": "2025-09-01T07:53:17Z",
                        }
                    }
                },
            },
            400: {"description": "Invalid request data", "model": BadRequestResponse},
            404: {"description": "Memory not found", "model": NotFoundResponse},
            422: {"description": "Validation error in payload", "model": ValidationErrorResponse},
            500: {"description": "Failed to update memory", "model": InternalServerErrorResponse},
        },
    )
    async def update_memory(
        request: Request,
        payload: UserMemoryCreateSchema,
        memory_id: str = Path(description="Memory ID to update"),
        db_id: Optional[str] = Query(default=None, description="Database ID to use for update"),
        table: Optional[str] = Query(default=None, description="Table to use for update"),
    ) -> UserMemorySchema:
        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id
            payload.user_id = user_id

        if payload.user_id is None:
            raise HTTPException(status_code=400, detail="User ID is required")

        db = await get_db(dbs, db_id, table)

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.update_memory(
                memory_id=memory_id,
                user_id=payload.user_id,
                memory=payload.memory,
                topics=payload.topics or [],
                db_id=db_id,
                table=table,
                headers=headers,
            )

        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            user_memory = await db.upsert_user_memory(
                memory=UserMemory(
                    memory_id=memory_id,
                    memory=payload.memory,
                    topics=payload.topics or [],
                    user_id=payload.user_id,
                ),
                deserialize=False,
            )
        else:
            user_memory = db.upsert_user_memory(
                memory=UserMemory(
                    memory_id=memory_id,
                    memory=payload.memory,
                    topics=payload.topics or [],
                    user_id=payload.user_id,
                ),
                deserialize=False,
            )
        if not user_memory:
            raise HTTPException(status_code=500, detail="Failed to update memory")

        return UserMemorySchema.from_dict(user_memory)  # type: ignore

    @router.get(
        "/user_memory_stats",
        response_model=PaginatedResponse[UserStatsSchema],
        status_code=200,
        operation_id="get_user_memory_stats",
        summary="Get User Memory Statistics",
        description=(
            "Retrieve paginated statistics about memory usage by user. "
            "Provides insights into user engagement and memory distribution across users."
        ),
        responses={
            200: {
                "description": "User memory statistics retrieved successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "data": [
                                {
                                    "user_id": "123",
                                    "total_memories": 3,
                                    "last_memory_updated_at": "2025-09-01T07:53:17Z",
                                }
                            ]
                        }
                    }
                },
            },
            500: {"description": "Failed to retrieve user statistics", "model": InternalServerErrorResponse},
        },
    )
    async def get_user_memory_stats(
        request: Request,
        limit: Optional[int] = Query(default=20, description="Number of user statistics to return per page", ge=1),
        page: Optional[int] = Query(default=1, description="Page number for pagination", ge=0),
        user_id: Optional[str] = Query(default=None, description="User ID to filter statistics for"),
        db_id: Optional[str] = Query(default=None, description="Database ID to query statistics from"),
        table: Optional[str] = Query(default=None, description="Table to query statistics from"),
    ) -> PaginatedResponse[UserStatsSchema]:
        db = await get_db(dbs, db_id, table)

        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            user_id = request.state.user_id

        if isinstance(db, RemoteDb):
            auth_token = get_auth_token_from_request(request)
            headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
            return await db.get_user_memory_stats(
                limit=limit,
                page=page,
                db_id=db_id,
                table=table,
                headers=headers,
            )

        try:
            # Ensure limit and page are integers
            limit = int(limit) if limit is not None else 20
            page = int(page) if page is not None else 1
            if isinstance(db, AsyncBaseDb):
                db = cast(AsyncBaseDb, db)
                user_stats, total_count = await db.get_user_memory_stats(
                    limit=limit,
                    page=page,
                    user_id=user_id,
                )
            else:
                user_stats, total_count = db.get_user_memory_stats(
                    limit=limit,
                    page=page,
                    user_id=user_id,
                )
            return PaginatedResponse(
                data=[UserStatsSchema.from_dict(stats) for stats in user_stats],
                meta=PaginationInfo(
                    page=page,
                    limit=limit,
                    total_count=total_count,
                    total_pages=(total_count + limit - 1) // limit if limit is not None and limit > 0 else 0,
                ),
            )

        except Exception as e:
            raise HTTPException(status_code=500, detail=f"Failed to get user statistics: {str(e)}")

    @router.post(
        "/optimize-memories",
        response_model=OptimizeMemoriesResponse,
        status_code=200,
        operation_id="optimize_memories",
        summary="Optimize User Memories",
        description=(
            "Optimize all memories for a given user using the default summarize strategy. "
            "This operation combines all memories into a single comprehensive summary, "
            "achieving maximum token reduction while preserving all key information. "
            "To use a custom model, specify the model parameter in 'provider:model_id' format "
            "(e.g., 'openai:gpt-4o-mini', 'anthropic:claude-3-5-sonnet-20241022'). "
            "If not specified, uses MemoryManager's default model (gpt-4o). "
            "Set apply=false to preview optimization results without saving to database."
        ),
        responses={
            200: {
                "description": "Memories optimized successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "memories": [
                                {
                                    "memory_id": "f9361a69-2997-40c7-ae4e-a5861d434047",
                                    "memory": "User has a 3-year-old golden retriever named Max who loves fetch and walks. Lives in San Francisco's Mission district, works as a product manager in tech. Enjoys hiking Bay Area trails, trying new restaurants (especially Japanese, Thai, Mexican), and learning piano for 1.5 years.",
                                    "topics": ["pets", "location", "work", "hobbies", "food_preferences"],
                                    "user_id": "user2",
                                    "updated_at": "2025-11-18T10:30:00Z",
                                }
                            ],
                            "memories_before": 4,
                            "memories_after": 1,
                            "tokens_before": 450,
                            "tokens_after": 180,
                            "tokens_saved": 270,
                            "reduction_percentage": 60.0,
                        }
                    }
                },
            },
            400: {
                "description": "Bad request - User ID is required or invalid model string format",
                "model": BadRequestResponse,
            },
            404: {"description": "No memories found for user", "model": NotFoundResponse},
            500: {"description": "Failed to optimize memories", "model": InternalServerErrorResponse},
        },
    )
    async def optimize_memories(
        http_request: Request,
        request: OptimizeMemoriesRequest,
        db_id: Optional[str] = Query(default=None, description="Database ID to use for optimization"),
        table: Optional[str] = Query(default=None, description="Table to use for optimization"),
    ) -> OptimizeMemoriesResponse:
        """Optimize user memories using the default summarize strategy."""
        from agno.memory import MemoryManager
        from agno.memory.strategies.types import MemoryOptimizationStrategyType

        if hasattr(http_request.state, "user_id") and http_request.state.user_id is not None:
            request.user_id = http_request.state.user_id

        try:
            # Get database instance
            db = await get_db(dbs, db_id, table)

            if isinstance(db, RemoteDb):
                auth_token = get_auth_token_from_request(http_request)
                headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else None
                return await db.optimize_memories(
                    user_id=request.user_id,
                    model=request.model,
                    apply=request.apply,
                    db_id=db_id,
                    table=table,
                    headers=headers,
                )

            # Create memory manager with optional model
            if request.model:
                try:
                    model_instance = get_model(request.model)
                except ValueError as e:
                    raise HTTPException(status_code=400, detail=str(e))
                memory_manager = MemoryManager(model=model_instance, db=db)
            else:
                # No model specified - use MemoryManager's default
                memory_manager = MemoryManager(db=db)

            # Get current memories to count tokens before optimization
            if isinstance(db, AsyncBaseDb):
                memories_before = await memory_manager.aget_user_memories(user_id=request.user_id)
            else:
                memories_before = memory_manager.get_user_memories(user_id=request.user_id)

            if not memories_before:
                raise HTTPException(status_code=404, detail=f"No memories found for user {request.user_id}")

            # Count tokens before optimization
            from agno.memory.strategies.summarize import SummarizeStrategy

            strategy = SummarizeStrategy()
            tokens_before = strategy.count_tokens(memories_before)
            memories_before_count = len(memories_before)

            # Optimize memories with default SUMMARIZE strategy
            if isinstance(db, AsyncBaseDb):
                optimized_memories = await memory_manager.aoptimize_memories(
                    user_id=request.user_id,
                    strategy=MemoryOptimizationStrategyType.SUMMARIZE,
                    apply=request.apply,
                )
            else:
                optimized_memories = memory_manager.optimize_memories(
                    user_id=request.user_id,
                    strategy=MemoryOptimizationStrategyType.SUMMARIZE,
                    apply=request.apply,
                )

            # Count tokens after optimization
            tokens_after = strategy.count_tokens(optimized_memories)
            memories_after_count = len(optimized_memories)

            # Calculate statistics
            tokens_saved = tokens_before - tokens_after
            reduction_percentage = (tokens_saved / tokens_before * 100.0) if tokens_before > 0 else 0.0

            # Convert to schema objects
            optimized_memory_schemas = [
                UserMemorySchema(
                    memory_id=mem.memory_id or "",
                    memory=mem.memory or "",
                    topics=mem.topics,
                    agent_id=mem.agent_id,
                    team_id=mem.team_id,
                    user_id=mem.user_id,
                    updated_at=mem.updated_at,
                )
                for mem in optimized_memories
            ]

            return OptimizeMemoriesResponse(
                memories=optimized_memory_schemas,
                memories_before=memories_before_count,
                memories_after=memories_after_count,
                tokens_before=tokens_before,
                tokens_after=tokens_after,
                tokens_saved=tokens_saved,
                reduction_percentage=reduction_percentage,
            )

        except HTTPException:
            raise
        except Exception as e:
            logger.error(f"Failed to optimize memories for user {request.user_id}: {str(e)}")
            raise HTTPException(status_code=500, detail=f"Failed to optimize memories: {str(e)}")

    return router


def parse_topics(
    topics: Optional[List[str]] = Query(
        default=None,
        description="Comma-separated list of topics to filter by",
        examples=["preferences,technical,communication_style"],
    ),
) -> Optional[List[str]]:
    """Parse comma-separated topics into a list for filtering memories by topic."""
    if not topics:
        return None

    try:
        return [topic.strip() for topic in topics[0].split(",") if topic.strip()]

    except Exception as e:
        raise HTTPException(status_code=422, detail=f"Invalid topics format: {e}")



"""
Learning Store Protocol
=======================
Defines the interface that all learning stores must implement.

This protocol enables:
- Consistent API across different learning types
- Easy addition of custom stores
- Type safety with Protocol typing
"""

from typing import Any, Callable, List, Optional, Protocol, runtime_checkable


@runtime_checkable
class LearningStore(Protocol):
    """Protocol that all learning stores must implement.

    A learning store handles one type of learning (user profile, session context,
    learned knowledge, etc.) and provides methods for:

    - recall: Retrieve relevant data for the current context
    - process: Extract and save learnings from conversations
    - build_context: Format data for inclusion in agent prompts
    - get_tools: Provide tools for agent interaction
    """

    @property
    def learning_type(self) -> str:
        """Unique identifier for this learning type.

        Used for storage keys and logging.

        Returns:
            String identifier (e.g., "user_profile", "session_context")
        """
        ...

    @property
    def schema(self) -> Any:
        """Schema class used for this learning type.

        Returns:
            The dataclass or schema class for this learning type.
        """
        ...

    def recall(self, **kwargs) -> Optional[Any]:
        """Retrieve relevant data for the current context.

        Args:
            **kwargs: Context including user_id, session_id, message, etc.

        Returns:
            Retrieved data (schema instance, list, or None if not found).
        """
        ...

    async def arecall(self, **kwargs) -> Optional[Any]:
        """Async version of recall."""
        ...

    def process(self, messages: List[Any], **kwargs) -> None:
        """Extract and save learnings from messages.

        Called after a conversation to extract learnings.

        Args:
            messages: Conversation messages to analyze.
            **kwargs: Context including user_id, session_id, etc.
        """
        ...

    async def aprocess(self, messages: List[Any], **kwargs) -> None:
        """Async version of process."""
        ...

    def build_context(self, data: Any) -> str:
        """Build context string for agent prompts.

        Formats the recalled data into a string that can be
        injected into the agent's system prompt.

        Args:
            data: Data returned from recall().

        Returns:
            Formatted context string, or empty string if no data.
        """
        ...

    def get_tools(self, **kwargs) -> List[Callable]:
        """Get tools to expose to the agent.

        Returns callable tools that the agent can use to interact
        with this learning type (e.g., update_user_memory, search_learnings).

        Args:
            **kwargs: Context including user_id, session_id, etc.

        Returns:
            List of callable tools, or empty list if no tools.
        """
        ...

    async def aget_tools(self, **kwargs) -> List[Callable]:
        """Async version of get_tools."""
        ...

    @property
    def was_updated(self) -> bool:
        """Check if the store was updated in the last operation.

        Returns:
            True if data was saved/updated, False otherwise.
        """
        ...



import csv
from os import getenv
from typing import Any, Dict, List, Optional

try:
    import redshift_connector
    from redshift_connector import Connection
except ImportError:
    raise ImportError("`redshift_connector` not installed. Please install using `pip install redshift-connector`.")

from agno.tools import Toolkit
from agno.utils.log import log_debug, log_error, log_info


class RedshiftTools(Toolkit):
    """
    A toolkit for interacting with Amazon Redshift databases.

    Supports these authentication methods:
    - Standard username and password authentication
    - IAM authentication with AWS profile
    - IAM authentication with AWS credentials

    Args:
        host (Optional[str]): Redshift cluster endpoint hostname. Falls back to REDSHIFT_HOST env var.
        port (int): Redshift cluster port number. Default is 5439.
        database (Optional[str]): Database name to connect to. Falls back to REDSHIFT_DATABASE env var.
        user (Optional[str]): Username for standard authentication.
        password (Optional[str]): Password for standard authentication.
        iam (bool): Enable IAM authentication. Default is False.
        cluster_identifier (Optional[str]): Redshift cluster identifier for IAM auth with provisioned clusters. Falls back to REDSHIFT_CLUSTER_IDENTIFIER env var.
        region (Optional[str]): AWS region for IAM credential retrieval. Falls back to AWS_REGION or AWS_DEFAULT_REGION env vars.
        db_user (Optional[str]): Database user for IAM auth with provisioned clusters. Falls back to REDSHIFT_DB_USER env var.
        access_key_id (Optional[str]): AWS access key ID for IAM auth. Falls back to AWS_ACCESS_KEY_ID env var.
        secret_access_key (Optional[str]): AWS secret access key for IAM auth. Falls back to AWS_SECRET_ACCESS_KEY env var.
        session_token (Optional[str]): AWS session token for temporary credentials. Falls back to AWS_SESSION_TOKEN env var.
        profile (Optional[str]): AWS profile name for IAM auth. Falls back to AWS_PROFILE env var.
        ssl (bool): Enable SSL connection. Default is True.
        table_schema (str): Default schema for table operations. Default is "public".
    """

    _requires_connect: bool = True

    def __init__(
        self,
        # Connection parameters
        host: Optional[str] = None,
        port: int = 5439,
        database: Optional[str] = None,
        # Standard authentication (username/password)
        user: Optional[str] = None,
        password: Optional[str] = None,
        # IAM Authentication
        iam: bool = False,
        cluster_identifier: Optional[str] = None,
        region: Optional[str] = None,
        db_user: Optional[str] = None,
        # AWS Credentials (for IAM auth)
        access_key_id: Optional[str] = None,
        secret_access_key: Optional[str] = None,
        session_token: Optional[str] = None,
        profile: Optional[str] = None,
        # Connection settings
        ssl: bool = True,
        table_schema: str = "public",
        **kwargs,
    ):
        # Connection parameters
        self.host: Optional[str] = host or getenv("REDSHIFT_HOST")
        self.port: int = port
        self.database: Optional[str] = database or getenv("REDSHIFT_DATABASE")

        # Standard authentication
        self.user: Optional[str] = user
        self.password: Optional[str] = password

        # IAM authentication parameters
        self.iam: bool = iam
        self.cluster_identifier: Optional[str] = cluster_identifier or getenv("REDSHIFT_CLUSTER_IDENTIFIER")
        self.region: Optional[str] = region or getenv("AWS_REGION") or getenv("AWS_DEFAULT_REGION")
        self.db_user: Optional[str] = db_user or getenv("REDSHIFT_DB_USER")

        # AWS credentials
        self.access_key_id: Optional[str] = access_key_id or getenv("AWS_ACCESS_KEY_ID")
        self.secret_access_key: Optional[str] = secret_access_key or getenv("AWS_SECRET_ACCESS_KEY")
        self.session_token: Optional[str] = session_token or getenv("AWS_SESSION_TOKEN")
        self.profile: Optional[str] = profile or getenv("AWS_PROFILE")

        # Connection settings
        self.ssl: bool = ssl
        self.table_schema: str = table_schema

        # Connection instance
        self._connection: Optional[Connection] = None

        tools: List[Any] = [
            self.show_tables,
            self.describe_table,
            self.summarize_table,
            self.inspect_query,
            self.run_query,
            self.export_table_to_path,
        ]

        super().__init__(name="redshift_tools", tools=tools, **kwargs)

    def connect(self) -> Connection:
        """
        Establish a connection to the Redshift database.

        Returns:
            The database connection object.

        Raises:
            redshift_connector.Error: If connection fails.
        """
        if self._connection is not None:
            log_debug("Connection already established, reusing existing connection")
            return self._connection

        log_info("Establishing connection to Redshift")
        self._connection = redshift_connector.connect(**self._get_connection_kwargs())
        return self._connection

    def close(self) -> None:
        """
        Close the database connection if it exists.
        """
        if self._connection is not None:
            log_info("Closing Redshift connection")
            try:
                self._connection.close()
            except Exception:
                pass  # Connection might already be closed
            self._connection = None

    @property
    def is_connected(self) -> bool:
        """Check if a connection is currently established."""
        return self._connection is not None

    def _ensure_connection(self) -> Connection:
        """
        Ensure a connection exists, creating one if necessary.

        Returns:
            The database connection object.
        """
        if self._connection is None:
            return self.connect()
        return self._connection

    def _get_connection_kwargs(self) -> Dict[str, Any]:
        """Build connection kwargs from instance."""
        connection_kwargs: Dict[str, Any] = {}

        # Common connection parameters
        if self.host:
            connection_kwargs["host"] = self.host
        if self.port:
            connection_kwargs["port"] = self.port
        if self.database:
            connection_kwargs["database"] = self.database
        connection_kwargs["ssl"] = self.ssl

        # IAM Authentication
        if self.iam:
            connection_kwargs["iam"] = True

            # For provisioned clusters (not serverless)
            if self.cluster_identifier:
                connection_kwargs["cluster_identifier"] = self.cluster_identifier
                # db_user required for provisioned clusters with IAM
                if self.db_user:
                    connection_kwargs["db_user"] = self.db_user

            # Region for IAM credential retrieval
            if self.region:
                connection_kwargs["region"] = self.region

            # AWS credentials - either profile or explicit
            if self.profile:
                connection_kwargs["profile"] = self.profile
            else:
                # Explicit AWS credentials
                if self.access_key_id:
                    connection_kwargs["access_key_id"] = self.access_key_id
                if self.secret_access_key:
                    connection_kwargs["secret_access_key"] = self.secret_access_key
                if self.session_token:
                    connection_kwargs["session_token"] = self.session_token

        else:
            # Standard username/password authentication
            if self.user:
                connection_kwargs["user"] = self.user
            if self.password:
                connection_kwargs["password"] = self.password

        return connection_kwargs

    def _execute_query(self, query: str, params: Optional[tuple] = None) -> str:
        try:
            connection = self._ensure_connection()
            with connection.cursor() as cursor:
                log_debug("Running Redshift query")
                cursor.execute(query, params)

                if cursor.description is None:
                    return "Query executed successfully."

                columns = [desc[0] for desc in cursor.description]
                rows = cursor.fetchall()

                if not rows:
                    return f"Query returned no results.\nColumns: {', '.join(columns)}"

                header = ",".join(columns)
                data_rows = [",".join(map(str, row)) for row in rows]
                return f"{header}\n" + "\n".join(data_rows)

        except redshift_connector.Error as e:
            log_error(f"Database error: {e}")
            if self._connection:
                try:
                    self._connection.rollback()
                except Exception:
                    pass  # Connection might be closed
            return f"Error executing query: {e}"
        except Exception as e:
            log_error(f"An unexpected error occurred: {e}")
            return f"An unexpected error occurred: {e}"

    def show_tables(self) -> str:
        """Lists all tables in the configured schema."""

        stmt = "SELECT table_name FROM information_schema.tables WHERE table_schema = %s;"
        return self._execute_query(stmt, (self.table_schema,))

    def describe_table(self, table: str) -> str:
        """
        Provides the schema (column name, data type, is nullable) for a given table.

        Args:
            table: The name of the table to describe.

        Returns:
            A string describing the table's columns and data types.
        """
        stmt = """
            SELECT column_name, data_type, is_nullable
            FROM information_schema.columns
            WHERE table_schema = %s AND table_name = %s;
        """
        return self._execute_query(stmt, (self.table_schema, table))

    def summarize_table(self, table: str) -> str:
        """
        Computes and returns key summary statistics for a table's columns.

        Args:
            table: The name of the table to summarize.

        Returns:
            A string containing a summary of the table.
        """
        try:
            connection = self._ensure_connection()
            with connection.cursor() as cursor:
                # First, get column information using a parameterized query
                schema_query = """
                    SELECT column_name, data_type
                    FROM information_schema.columns
                    WHERE table_schema = %s AND table_name = %s;
                """
                cursor.execute(schema_query, (self.table_schema, table))
                columns = cursor.fetchall()
                if not columns:
                    return f"Error: Table '{table}' not found in schema '{self.table_schema}'."

                summary_parts = [f"Summary for table: {table}\n"]

                # Redshift uses schema.table format for fully qualified names
                full_table_name = f'"{self.table_schema}"."{table}"'

                for col in columns:
                    col_name = col[0]
                    data_type = col[1]

                    query = None
                    if any(
                        t in data_type.lower()
                        for t in [
                            "integer",
                            "int",
                            "bigint",
                            "smallint",
                            "numeric",
                            "decimal",
                            "real",
                            "double precision",
                            "float",
                        ]
                    ):
                        query = f"""
                            SELECT
                                COUNT(*) AS total_rows,
                                COUNT("{col_name}") AS non_null_rows,
                                MIN("{col_name}") AS min,
                                MAX("{col_name}") AS max,
                                AVG("{col_name}") AS average,
                                STDDEV("{col_name}") AS std_deviation
                            FROM {full_table_name};
                        """
                    elif any(t in data_type.lower() for t in ["char", "varchar", "text", "uuid"]):
                        query = f"""
                            SELECT
                                COUNT(*) AS total_rows,
                                COUNT("{col_name}") AS non_null_rows,
                                COUNT(DISTINCT "{col_name}") AS unique_values,
                                AVG(LEN("{col_name}")) as avg_length
                            FROM {full_table_name};
                        """

                    if query:
                        cursor.execute(query)
                        stats = cursor.fetchone()
                        summary_parts.append(f"\n--- Column: {col_name} (Type: {data_type}) ---")
                        if stats is not None:
                            stats_dict = dict(zip([desc[0] for desc in cursor.description], stats))
                            for key, value in stats_dict.items():
                                val_str = (
                                    f"{value:.2f}" if isinstance(value, float) and value is not None else str(value)
                                )
                                summary_parts.append(f"  {key}: {val_str}")
                        else:
                            summary_parts.append("  No statistics available")

                return "\n".join(summary_parts)

        except redshift_connector.Error as e:
            return f"Error summarizing table: {e}"

    def inspect_query(self, query: str) -> str:
        """
        Shows the execution plan for a SQL query (using EXPLAIN).

        Args:
            query: The SQL query to inspect.

        Returns:
            The query's execution plan.
        """
        return self._execute_query(f"EXPLAIN {query}")

    def export_table_to_path(self, table: str, path: str) -> str:
        """
        Exports a table's data to a local CSV file.

        Args:
            table: The name of the table to export.
            path: The local file path to save the file.

        Returns:
            A confirmation message with the file path.
        """
        log_debug(f"Exporting table {table} to {path}")

        full_table_name = f'"{self.table_schema}"."{table}"'
        stmt = f"SELECT * FROM {full_table_name};"

        try:
            connection = self._ensure_connection()
            with connection.cursor() as cursor:
                cursor.execute(stmt)

                if cursor.description is None:
                    return f"Error: Query returned no description for table '{table}'."

                columns = [desc[0] for desc in cursor.description]

                with open(path, "w", newline="", encoding="utf-8") as f:
                    writer = csv.writer(f)
                    writer.writerow(columns)
                    writer.writerows(cursor)

            return f"Successfully exported table '{table}' to '{path}'."
        except (redshift_connector.Error, IOError) as e:
            if self._connection:
                try:
                    self._connection.rollback()
                except Exception:
                    pass  # Connection might be closed
            return f"Error exporting table: {e}"

    def run_query(self, query: str) -> str:
        """
        Runs a read-only SQL query and returns the result.

        Args:
            query: The SQL query to run.

        Returns:
            The query result as a formatted string.
        """
        return self._execute_query(query)



import asyncio
import json
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from hashlib import md5
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union, cast

try:
    from chromadb import Client as ChromaDbClient
    from chromadb import PersistentClient as PersistentChromaDbClient
    from chromadb.api.client import ClientAPI
    from chromadb.api.models.Collection import Collection
    from chromadb.api.types import QueryResult

except ImportError:
    raise ImportError("The `chromadb` package is not installed. Please install it via `pip install chromadb`.")

from agno.filters import FilterExpr
from agno.knowledge.document import Document
from agno.knowledge.embedder import Embedder
from agno.knowledge.reranker.base import Reranker
from agno.utils.log import log_debug, log_error, log_info, log_warning, logger
from agno.vectordb.base import VectorDb
from agno.vectordb.distance import Distance
from agno.vectordb.search import SearchType


def reciprocal_rank_fusion(
    ranked_lists: List[List[Tuple[str, float]]],
    k: int = 60,
) -> List[Tuple[str, float]]:
    """
    Combine multiple ranked lists using Reciprocal Rank Fusion (RRF).

    RRF is a simple yet effective method for combining multiple rankings.
    The formula is: RRF(d) = sum(1 / (k + rank_i(d))) for each ranking i

    Args:
        ranked_lists: List of ranked results, each as [(doc_id, score), ...]
        k: RRF constant (default 60, as per original paper by Cormack et al.)

    Returns:
        Fused ranking as [(doc_id, rrf_score), ...] sorted by score descending
    """
    rrf_scores: Dict[str, float] = defaultdict(float)

    for ranked_list in ranked_lists:
        for rank, (doc_id, _) in enumerate(ranked_list, start=1):
            rrf_scores[doc_id] += 1.0 / (k + rank)

    sorted_results = sorted(rrf_scores.items(), key=lambda x: x[1], reverse=True)
    return sorted_results


class ChromaDb(VectorDb):
    """
    ChromaDb class for managing vector operations with ChromaDB.

    Args:
        collection: The name of the ChromaDB collection. If not provided, derived from 'name'.
        name: Name of the vector database. Also used as collection name if 'collection' is not provided.
        description: Description of the vector database.
        id: Unique identifier for this vector database instance.
        embedder: The embedder to use when embedding the document contents.
        distance: The distance metric to use when searching for documents.
        path: The path to store the ChromaDB data (for persistent client).
        persistent_client: Whether to use a persistent client.
        search_type: The search type to use when searching for documents.
            - SearchType.vector: Pure vector similarity search (default)
            - SearchType.keyword: Keyword-based search using document content
            - SearchType.hybrid: Combines vector + FTS with Reciprocal Rank Fusion
        hybrid_rrf_k: RRF (Reciprocal Rank Fusion) constant for hybrid search.
            Controls ranking smoothness - higher values give more weight to lower-ranked
            results, lower values make top results more dominant. Default is 60
            (per original RRF paper by Cormack et al.).
        reranker: The reranker to use when reranking documents.
        **kwargs: Additional arguments to pass to the ChromaDB client.
    """

    def __init__(
        self,
        collection: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        embedder: Optional[Embedder] = None,
        distance: Distance = Distance.cosine,
        path: str = "tmp/chromadb",
        persistent_client: bool = False,
        search_type: SearchType = SearchType.vector,
        hybrid_rrf_k: int = 60,
        reranker: Optional[Reranker] = None,
        **kwargs,
    ):
        # Derive collection from name if not provided
        if collection is None:
            if name is not None:
                # Sanitize name: lowercase and replace spaces with underscores
                collection = name.lower().replace(" ", "_")
            else:
                raise ValueError("Either 'collection' or 'name' must be provided.")

        # Dynamic ID generation based on unique identifiers
        if id is None:
            from agno.utils.string import generate_id

            seed = f"{path}#{collection}"
            id = generate_id(seed)

        # Initialize base class with name, description, and generated ID
        super().__init__(id=id, name=name, description=description)

        # Collection attributes
        self.collection_name: str = collection
        # Embedder for embedding the document contents
        if embedder is None:
            from agno.knowledge.embedder.openai import OpenAIEmbedder

            embedder = OpenAIEmbedder()
            log_debug("Embedder not provided, using OpenAIEmbedder as default.")
        self.embedder: Embedder = embedder
        # Distance metric
        self.distance: Distance = distance

        # Chroma client instance
        self._client: Optional[ClientAPI] = None

        # Chroma collection instance
        self._collection: Optional[Collection] = None

        # Persistent Chroma client instance
        self.persistent_client: bool = persistent_client
        self.path: str = path

        # Search type configuration
        self.search_type: SearchType = search_type
        self.hybrid_rrf_k: int = hybrid_rrf_k

        # Reranker instance
        self.reranker: Optional[Reranker] = reranker

        # Chroma client kwargs
        self.kwargs = kwargs

    def _flatten_metadata(self, metadata: Dict[str, Any]) -> Dict[str, Union[str, int, float, bool]]:
        """
        Flatten nested metadata to ChromaDB-compatible format.

        Args:
            metadata: Dictionary that may contain nested structures

        Returns:
            Flattened dictionary with only primitive values
        """
        flattened: Dict[str, Any] = {}

        def _flatten_recursive(obj: Any, prefix: str = "") -> None:
            if isinstance(obj, dict):
                if len(obj) == 0:
                    # Handle empty dictionaries by converting to JSON string
                    flattened[prefix] = json.dumps(obj)
                else:
                    for key, value in obj.items():
                        new_key = f"{prefix}.{key}" if prefix else key
                        _flatten_recursive(value, new_key)
            elif isinstance(obj, (list, tuple)):
                # Convert lists/tuples to JSON strings
                flattened[prefix] = json.dumps(obj)
            elif isinstance(obj, (str, int, float, bool)) or obj is None:
                if obj is not None:  # ChromaDB doesn't accept None values
                    flattened[prefix] = obj
            else:
                # Convert other complex types to JSON strings
                try:
                    flattened[prefix] = json.dumps(obj)
                except (TypeError, ValueError):
                    # If it can't be serialized, convert to string
                    flattened[prefix] = str(obj)

        _flatten_recursive(metadata)
        return flattened

    @property
    def client(self) -> ClientAPI:
        if self._client is None:
            if not self.persistent_client:
                log_debug("Creating Chroma Client")
                self._client = ChromaDbClient(
                    **self.kwargs,
                )
            elif self.persistent_client:
                log_debug("Creating Persistent Chroma Client")
                self._client = PersistentChromaDbClient(
                    path=self.path,
                    **self.kwargs,
                )
        return self._client

    def create(self) -> None:
        """Create the collection in ChromaDb."""
        if self.exists():
            log_debug(f"Collection already exists: {self.collection_name}")
            self._collection = self.client.get_collection(name=self.collection_name)
        else:
            log_debug(f"Creating collection: {self.collection_name}")
            self._collection = self.client.create_collection(
                name=self.collection_name, metadata={"hnsw:space": self.distance.value}
            )

    async def async_create(self) -> None:
        """Create the collection asynchronously by running in a thread."""
        await asyncio.to_thread(self.create)

    def name_exists(self, name: str) -> bool:
        """Check if a document with a given name exists in the collection.
        Args:
            name (str): Name of the document to check.
        Returns:
            bool: True if document exists, False otherwise."""
        if not self.client:
            logger.warning("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)
            result = collection.get(where=cast(Any, {"name": {"$eq": name}}), limit=1)
            return len(result.get("ids", [])) > 0
        except Exception as e:
            logger.error(f"Error checking name existence: {e}")
        return False

    async def async_name_exists(self, name: str) -> bool:
        """Check if a document with given name exists asynchronously."""
        return await asyncio.to_thread(self.name_exists, name)

    def insert(self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None) -> None:
        """Insert documents into the collection.

        Args:
            documents (List[Document]): List of documents to insert
            filters (Optional[Dict[str, Any]]): Filters to merge with document metadata
        """
        log_info(f"Inserting {len(documents)} documents")
        ids: List = []
        docs: List = []
        docs_embeddings: List = []
        docs_metadata: List = []

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        for document in documents:
            document.embed(embedder=self.embedder)
            cleaned_content = document.content.replace("\x00", "\ufffd")
            doc_id = md5(cleaned_content.encode()).hexdigest()

            # Handle metadata and filters
            metadata = document.meta_data or {}
            if filters:
                metadata.update(filters)

            # Add name, content_id to metadata
            if document.name is not None:
                metadata["name"] = document.name
            if document.content_id is not None:
                metadata["content_id"] = document.content_id

            metadata["content_hash"] = content_hash

            # Flatten metadata for ChromaDB compatibility
            flattened_metadata = self._flatten_metadata(metadata)

            docs_embeddings.append(document.embedding)
            docs.append(cleaned_content)
            ids.append(doc_id)
            docs_metadata.append(flattened_metadata)
            log_debug(f"Prepared document: {document.id} | {document.name} | {flattened_metadata}")

        if self._collection is None:
            logger.warning("Collection does not exist")
        else:
            if len(docs) > 0:
                self._collection.add(ids=ids, embeddings=docs_embeddings, documents=docs, metadatas=docs_metadata)
                log_debug(f"Committed {len(docs)} documents")

    async def async_insert(
        self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None
    ) -> None:
        """Insert documents asynchronously by running in a thread."""
        log_info(f"Async Inserting {len(documents)} documents")
        ids: List = []
        docs: List = []
        docs_embeddings: List = []
        docs_metadata: List = []

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        if self.embedder.enable_batch and hasattr(self.embedder, "async_get_embeddings_batch_and_usage"):
            # Use batch embedding when enabled and supported
            try:
                # Extract content from all documents
                doc_contents = [doc.content for doc in documents]

                # Get batch embeddings and usage
                embeddings, usages = await self.embedder.async_get_embeddings_batch_and_usage(doc_contents)

                # Process documents with pre-computed embeddings
                for j, doc in enumerate(documents):
                    try:
                        if j < len(embeddings):
                            doc.embedding = embeddings[j]
                            doc.usage = usages[j] if j < len(usages) else None
                    except Exception as e:
                        logger.error(f"Error assigning batch embedding to document '{doc.name}': {e}")

            except Exception as e:
                # Check if this is a rate limit error - don't fall back as it would make things worse
                error_str = str(e).lower()
                is_rate_limit = any(
                    phrase in error_str
                    for phrase in ["rate limit", "too many requests", "429", "trial key", "api calls / minute"]
                )

                if is_rate_limit:
                    logger.error(f"Rate limit detected during batch embedding. {e}")
                    raise e
                else:
                    logger.warning(f"Async batch embedding failed, falling back to individual embeddings: {e}")
                    # Fall back to individual embedding
                    embed_tasks = [doc.async_embed(embedder=self.embedder) for doc in documents]
                    await asyncio.gather(*embed_tasks, return_exceptions=True)
        else:
            # Use individual embedding
            try:
                embed_tasks = [document.async_embed(embedder=self.embedder) for document in documents]
                await asyncio.gather(*embed_tasks, return_exceptions=True)
            except Exception as e:
                logger.error(f"Error processing document: {e}")

        for document in documents:
            cleaned_content = document.content.replace("\x00", "\ufffd")
            # Include content_hash in ID to ensure uniqueness across different content hashes
            base_id = document.id or md5(cleaned_content.encode()).hexdigest()
            doc_id = md5(f"{base_id}_{content_hash}".encode()).hexdigest()

            # Handle metadata and filters
            metadata = document.meta_data or {}
            if filters:
                metadata.update(filters)

            # Add name, content_id to metadata
            if document.name is not None:
                metadata["name"] = document.name
            if document.content_id is not None:
                metadata["content_id"] = document.content_id

            metadata["content_hash"] = content_hash

            # Flatten metadata for ChromaDB compatibility
            flattened_metadata = self._flatten_metadata(metadata)

            docs_embeddings.append(document.embedding)
            docs.append(cleaned_content)
            ids.append(doc_id)
            docs_metadata.append(flattened_metadata)
            log_debug(f"Prepared document: {document.id} | {document.name} | {flattened_metadata}")

        if self._collection is None:
            logger.warning("Collection does not exist")
        else:
            if len(docs) > 0:
                self._collection.add(ids=ids, embeddings=docs_embeddings, documents=docs, metadatas=docs_metadata)
                log_debug(f"Committed {len(docs)} documents")

    def upsert_available(self) -> bool:
        """Check if upsert is available in ChromaDB."""
        return True

    def upsert(self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None) -> None:
        """Upsert documents into the collection.

        Args:
            documents (List[Document]): List of documents to upsert
            filters (Optional[Dict[str, Any]]): Filters to apply while upserting
        """
        try:
            if self.content_hash_exists(content_hash):
                self._delete_by_content_hash(content_hash)
            self._upsert(content_hash, documents, filters)
        except Exception as e:
            logger.error(f"Error upserting documents by content hash: {e}")
            raise

    def _upsert(self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None) -> None:
        """Upsert documents into the collection.

        Args:
            documents (List[Document]): List of documents to upsert
            filters (Optional[Dict[str, Any]]): Filters to apply while upserting
        """
        log_info(f"Upserting {len(documents)} documents")
        ids: List = []
        docs: List = []
        docs_embeddings: List = []
        docs_metadata: List = []

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        for document in documents:
            document.embed(embedder=self.embedder)
            cleaned_content = document.content.replace("\x00", "\ufffd")
            doc_id = md5(cleaned_content.encode()).hexdigest()

            # Handle metadata and filters
            metadata = document.meta_data or {}
            if filters:
                metadata.update(filters)

            # Add name, content_id to metadata
            if document.name is not None:
                metadata["name"] = document.name
            if document.content_id is not None:
                metadata["content_id"] = document.content_id

            metadata["content_hash"] = content_hash

            # Flatten metadata for ChromaDB compatibility
            flattened_metadata = self._flatten_metadata(metadata)

            docs_embeddings.append(document.embedding)
            docs.append(cleaned_content)
            ids.append(doc_id)
            docs_metadata.append(flattened_metadata)
            log_debug(f"Upserted document: {document.id} | {document.name} | {flattened_metadata}")

        if self._collection is None:
            logger.warning("Collection does not exist")
        else:
            if len(docs) > 0:
                self._collection.upsert(ids=ids, embeddings=docs_embeddings, documents=docs, metadatas=docs_metadata)
                log_debug(f"Committed {len(docs)} documents")

    async def _async_upsert(
        self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None
    ) -> None:
        """Upsert documents into the collection.

        Args:
            documents (List[Document]): List of documents to upsert
            filters (Optional[Dict[str, Any]]): Filters to apply while upserting
        """
        log_info(f"Async Upserting {len(documents)} documents")
        ids: List = []
        docs: List = []
        docs_embeddings: List = []
        docs_metadata: List = []

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        if self.embedder.enable_batch and hasattr(self.embedder, "async_get_embeddings_batch_and_usage"):
            # Use batch embedding when enabled and supported
            try:
                # Extract content from all documents
                doc_contents = [doc.content for doc in documents]

                # Get batch embeddings and usage
                embeddings, usages = await self.embedder.async_get_embeddings_batch_and_usage(doc_contents)

                # Process documents with pre-computed embeddings
                for j, doc in enumerate(documents):
                    try:
                        if j < len(embeddings):
                            doc.embedding = embeddings[j]
                            doc.usage = usages[j] if j < len(usages) else None
                    except Exception as e:
                        logger.error(f"Error assigning batch embedding to document '{doc.name}': {e}")

            except Exception as e:
                # Check if this is a rate limit error - don't fall back as it would make things worse
                error_str = str(e).lower()
                is_rate_limit = any(
                    phrase in error_str
                    for phrase in ["rate limit", "too many requests", "429", "trial key", "api calls / minute"]
                )

                if is_rate_limit:
                    logger.error(f"Rate limit detected during batch embedding. {e}")
                    raise e
                else:
                    logger.warning(f"Async batch embedding failed, falling back to individual embeddings: {e}")
                    # Fall back to individual embedding
                    embed_tasks = [doc.async_embed(embedder=self.embedder) for doc in documents]
                    await asyncio.gather(*embed_tasks, return_exceptions=True)
        else:
            # Use individual embedding
            embed_tasks = [document.async_embed(embedder=self.embedder) for document in documents]
            await asyncio.gather(*embed_tasks, return_exceptions=True)

        for document in documents:
            cleaned_content = document.content.replace("\x00", "\ufffd")
            # Include content_hash in ID to ensure uniqueness across different content hashes
            base_id = document.id or md5(cleaned_content.encode()).hexdigest()
            doc_id = md5(f"{base_id}_{content_hash}".encode()).hexdigest()

            # Handle metadata and filters
            metadata = document.meta_data or {}
            if filters:
                metadata.update(filters)

            # Add name, content_id to metadata
            if document.name is not None:
                metadata["name"] = document.name
            if document.content_id is not None:
                metadata["content_id"] = document.content_id

            metadata["content_hash"] = content_hash

            # Flatten metadata for ChromaDB compatibility
            flattened_metadata = self._flatten_metadata(metadata)

            docs_embeddings.append(document.embedding)
            docs.append(cleaned_content)
            ids.append(doc_id)
            docs_metadata.append(flattened_metadata)
            log_debug(f"Upserted document: {document.id} | {document.name} | {flattened_metadata}")

        if self._collection is None:
            logger.warning("Collection does not exist")
        else:
            if len(docs) > 0:
                self._collection.upsert(ids=ids, embeddings=docs_embeddings, documents=docs, metadatas=docs_metadata)
                log_debug(f"Committed {len(docs)} documents")

    async def async_upsert(
        self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None
    ) -> None:
        """Upsert documents asynchronously by running in a thread."""
        try:
            if self.content_hash_exists(content_hash):
                self._delete_by_content_hash(content_hash)
            await self._async_upsert(content_hash, documents, filters)
        except Exception as e:
            logger.error(f"Error upserting documents by content hash: {e}")
            raise

    def search(
        self, query: str, limit: int = 5, filters: Optional[Union[Dict[str, Any], List[FilterExpr]]] = None
    ) -> List[Document]:
        """Search the collection for a query.

        Args:
            query (str): Query to search for.
            limit (int): Number of results to return.
            filters (Optional[Union[Dict[str, Any], List[FilterExpr]]]): Filters to apply while searching.
                Supports ChromaDB's filtering operators:
                - $eq, $ne: Equality/Inequality
                - $gt, $gte, $lt, $lte: Numeric comparisons
                - $in, $nin: List inclusion/exclusion
                - $and, $or: Logical operators
        Returns:
            List[Document]: List of search results.
        """
        if isinstance(filters, list):
            log_warning("Filter Expressions are not yet supported in ChromaDB. No filters will be applied.")
            filters = None

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        # Route to appropriate search method based on search_type
        if self.search_type == SearchType.vector:
            search_results = self._vector_search(query, limit, filters)
        elif self.search_type == SearchType.keyword:
            search_results = self._keyword_search(query, limit, filters)
        elif self.search_type == SearchType.hybrid:
            search_results = self._hybrid_search(query, limit, filters)
        else:
            logger.error(f"Invalid search type '{self.search_type}'.")
            return []

        if self.reranker and search_results:
            try:
                search_results = self.reranker.rerank(query=query, documents=search_results)
            except Exception as e:
                log_warning(f"Reranker failed, returning unranked results: {e}")

        log_info(f"Found {len(search_results)} documents")
        return search_results

    def _vector_search(self, query: str, limit: int = 5, filters: Optional[Dict[str, Any]] = None) -> List[Document]:
        """Perform pure vector similarity search.

        Args:
            query (str): Query to search for.
            limit (int): Number of results to return.
            filters (Optional[Dict[str, Any]]): Metadata filters to apply.

        Returns:
            List[Document]: List of search results.
        """
        query_embedding = self.embedder.get_embedding(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        # Convert simple filters to ChromaDB's format if needed
        where_filter = self._convert_filters(filters) if filters else None

        result: QueryResult = self._collection.query(
            query_embeddings=query_embedding,
            n_results=limit,
            where=where_filter,
            include=["metadatas", "documents", "embeddings", "distances", "uris"],
        )

        return self._build_search_results(result)

    def _keyword_search(self, query: str, limit: int = 5, filters: Optional[Dict[str, Any]] = None) -> List[Document]:
        """Perform keyword-based search using document content filtering.

        This uses ChromaDB's where_document filter with $contains operator
        for basic full-text search functionality.

        Args:
            query (str): Query to search for (keywords to match in document content).
            limit (int): Number of results to return.
            filters (Optional[Dict[str, Any]]): Metadata filters to apply.

        Returns:
            List[Document]: List of search results.
        """
        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        # Convert simple filters to ChromaDB's format if needed
        where_filter = self._convert_filters(filters) if filters else None

        # Get first significant word for $contains filter
        query_words = query.split()
        if not query_words:
            return []

        # Use where_document to filter by document content
        where_document: Dict[str, Any] = {"$contains": query_words[0]}

        try:
            # Get documents matching the keyword filter
            result = self._collection.get(
                where=where_filter,
                where_document=cast(Any, where_document),
                limit=limit,
                include=["metadatas", "documents", "embeddings"],
            )

            return self._build_get_results(cast(Dict[str, Any], result), query)
        except Exception as e:
            logger.error(f"Error in keyword search: {e}")
            return []

    def _hybrid_search(self, query: str, limit: int = 5, filters: Optional[Dict[str, Any]] = None) -> List[Document]:
        """Perform hybrid search combining vector similarity with full-text search using RRF.

        This method combines:
        1. Dense vector similarity search (semantic search)
        2. Full-text search (keyword/lexical search)

        Results are fused using Reciprocal Rank Fusion (RRF) for optimal ranking.

        Args:
            query (str): Query to search for.
            limit (int): Number of results to return.
            filters (Optional[Dict[str, Any]]): Metadata filters to apply.

        Returns:
            List[Document]: List of search results with RRF-fused ranking.
        """
        query_embedding = self.embedder.get_embedding(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []

        if not self._collection:
            self._collection = self.client.get_collection(name=self.collection_name)

        # Convert simple filters to ChromaDB's format if needed
        where_filter = self._convert_filters(filters) if filters else None

        # Fetch more candidates than needed for better fusion
        fetch_k = min(limit * 3, 100)

        def dense_vector_similarity_search() -> List[Tuple[str, float]]:
            """Dense vector similarity search."""
            try:
                results = self._collection.query(  # type: ignore
                    query_embeddings=query_embedding,
                    n_results=fetch_k,
                    where=where_filter,
                    include=["documents", "metadatas", "distances"],
                )

                ranked: List[Tuple[str, float]] = []
                if results.get("ids") and results["ids"][0]:
                    for i, doc_id in enumerate(results["ids"][0]):
                        distance = results["distances"][0][i] if results.get("distances") else 0  # type: ignore
                        # Convert distance to similarity score (lower distance = higher score)
                        score = 1.0 / (1.0 + distance)
                        ranked.append((doc_id, score))
                return ranked
            except Exception as e:
                log_error(f"Error in vector search component: {e}")
                return []

        def fts_search() -> List[Tuple[str, float]]:
            """Full-text search using ChromaDB's where_document filter."""
            try:
                query_words = query.split()
                if not query_words:
                    return []

                # Use first word for $contains filter
                fts_where_document: Dict[str, Any] = {"$contains": query_words[0]}

                results = self._collection.query(  # type: ignore
                    query_embeddings=query_embedding,
                    n_results=fetch_k,
                    where=where_filter,
                    where_document=cast(Any, fts_where_document),
                    include=["documents", "metadatas", "distances"],
                )

                ranked: List[Tuple[str, float]] = []
                if results.get("ids") and results["ids"][0]:
                    for i, doc_id in enumerate(results["ids"][0]):
                        # Score based on term overlap (simple BM25-like scoring)
                        doc = results["documents"][0][i] if results.get("documents") else ""  # type: ignore
                        query_terms = set(query.lower().split())
                        doc_terms = set(doc.lower().split()) if doc else set()
                        overlap = len(query_terms & doc_terms)
                        score = overlap / max(len(query_terms), 1)
                        ranked.append((doc_id, score))

                # Sort by score descending
                ranked.sort(key=lambda x: x[1], reverse=True)
                return ranked
            except Exception as e:
                log_error(f"Error in FTS search component: {e}")
                return []

        # Execute searches in parallel for better performance
        with ThreadPoolExecutor(max_workers=2) as executor:
            vector_future = executor.submit(dense_vector_similarity_search)
            fts_future = executor.submit(fts_search)

            vector_results = vector_future.result()
            fts_results = fts_future.result()

        # Apply RRF fusion
        fused_ranking = reciprocal_rank_fusion(
            [vector_results, fts_results],
            k=self.hybrid_rrf_k,
        )

        # Get top IDs from fused ranking
        top_ids = [doc_id for doc_id, _ in fused_ranking[:limit]]

        if not top_ids:
            return []

        # Fetch full document data for top results
        try:
            full_results = self._collection.get(
                ids=top_ids,
                include=["documents", "metadatas", "embeddings"],
            )
        except Exception as e:
            log_error(f"Error fetching full results: {e}")
            return []

        # Build lookup dict for results
        doc_lookup: Dict[str, Dict[str, Any]] = {}
        result_ids = full_results.get("ids", [])
        result_docs = full_results.get("documents")
        result_metas = full_results.get("metadatas")
        result_embeds = full_results.get("embeddings")

        for i, doc_id in enumerate(result_ids if result_ids is not None else []):
            doc_lookup[doc_id] = {
                "document": result_docs[i] if result_docs is not None and i < len(result_docs) else None,
                "metadata": result_metas[i] if result_metas is not None and i < len(result_metas) else None,
                "embedding": result_embeds[i] if result_embeds is not None and i < len(result_embeds) else None,
            }

        # Build final results in fused ranking order
        search_results: List[Document] = []
        rrf_scores = dict(fused_ranking)

        for doc_id in top_ids:
            if doc_id not in doc_lookup:
                continue

            doc_data = doc_lookup[doc_id]
            doc_metadata = dict(doc_data["metadata"]) if doc_data["metadata"] else {}

            # Add RRF score to metadata
            doc_metadata["rrf_score"] = rrf_scores.get(doc_id, 0.0)

            # Extract the fields we added to metadata
            name_val = doc_metadata.pop("name", None)
            content_id_val = doc_metadata.pop("content_id", None)

            # Convert types to match Document constructor expectations
            name = str(name_val) if name_val is not None and not isinstance(name_val, str) else name_val
            content_id = (
                str(content_id_val)
                if content_id_val is not None and not isinstance(content_id_val, str)
                else content_id_val
            )
            content = str(doc_data["document"]) if doc_data["document"] is not None else ""

            # Process embedding
            embedding = None
            if doc_data["embedding"] is not None:
                embed_data = doc_data["embedding"]
                if hasattr(embed_data, "tolist") and callable(getattr(embed_data, "tolist", None)):
                    try:
                        embedding = list(cast(Any, embed_data).tolist())
                    except (AttributeError, TypeError):
                        embedding = list(embed_data) if isinstance(embed_data, (list, tuple)) else None
                elif isinstance(embed_data, (list, tuple)):
                    embedding = [float(x) for x in embed_data if isinstance(x, (int, float))]

            search_results.append(
                Document(
                    id=doc_id,
                    name=name,
                    meta_data=doc_metadata,
                    content=content,
                    embedding=embedding,
                    content_id=content_id,
                )
            )

        return search_results

    def _build_search_results(self, result: QueryResult) -> List[Document]:
        """Build Document list from ChromaDB QueryResult.

        Args:
            result: The QueryResult from ChromaDB query.

        Returns:
            List[Document]: List of Document objects.
        """
        search_results: List[Document] = []

        ids_list = result.get("ids", [[]])  # type: ignore
        metadata_list = result.get("metadatas", [[{}]])  # type: ignore
        documents_list = result.get("documents", [[]])  # type: ignore
        embeddings_list = result.get("embeddings")  # type: ignore
        distances_list = result.get("distances", [[]])  # type: ignore

        # Check if we have valid results - handle numpy arrays carefully
        if ids_list is None or len(ids_list) == 0:
            return search_results
        if metadata_list is None or len(metadata_list) == 0:
            return search_results
        if documents_list is None or len(documents_list) == 0:
            return search_results
        if distances_list is None or len(distances_list) == 0:
            return search_results

        ids = ids_list[0]
        metadata = [dict(m) if m else {} for m in metadata_list[0]]  # Convert to mutable dicts
        documents = documents_list[0]

        # Handle embeddings - may be None or numpy array
        embeddings_raw: Any = []
        if embeddings_list is not None:
            try:
                if len(embeddings_list) > 0:
                    embeddings_raw = embeddings_list[0]
            except (TypeError, ValueError):
                # numpy array truth value issue - try direct access
                try:
                    embeddings_raw = embeddings_list[0]
                except Exception:
                    embeddings_raw = []

        embeddings = []
        for e in embeddings_raw:
            if hasattr(e, "tolist") and callable(getattr(e, "tolist", None)):
                try:
                    embeddings.append(list(cast(Any, e).tolist()))
                except (AttributeError, TypeError):
                    embeddings.append(list(e) if isinstance(e, (list, tuple)) else [])
            elif isinstance(e, (list, tuple)):
                embeddings.append([float(x) for x in e if isinstance(x, (int, float))])
            elif isinstance(e, (int, float)):
                embeddings.append([float(e)])
            else:
                embeddings.append([])

        distances = distances_list[0] if len(distances_list) > 0 else []

        for idx, distance in enumerate(distances):
            if idx < len(metadata):
                metadata[idx]["distances"] = distance

        try:
            for idx, (id_, doc_metadata, document) in enumerate(zip(ids, metadata, documents)):
                # Extract the fields we added to metadata
                name_val = doc_metadata.pop("name", None)
                content_id_val = doc_metadata.pop("content_id", None)

                # Convert types to match Document constructor expectations
                name = str(name_val) if name_val is not None and not isinstance(name_val, str) else name_val
                content_id = (
                    str(content_id_val)
                    if content_id_val is not None and not isinstance(content_id_val, str)
                    else content_id_val
                )
                content = str(document) if document is not None else ""
                embedding = embeddings[idx] if idx < len(embeddings) else None

                search_results.append(
                    Document(
                        id=id_,
                        name=name,
                        meta_data=doc_metadata,
                        content=content,
                        embedding=embedding,
                        content_id=content_id,
                    )
                )
        except Exception as e:
            log_error(f"Error building search results: {e}")

        return search_results

    def _build_get_results(self, result: Dict[str, Any], query: str = "") -> List[Document]:
        """Build Document list from ChromaDB GetResult.

        Args:
            result: The GetResult from ChromaDB get.
            query: The original query for scoring.

        Returns:
            List[Document]: List of Document objects.
        """
        search_results: List[Document] = []

        ids = result.get("ids", [])
        metadatas = result.get("metadatas", [])
        documents = result.get("documents", [])
        embeddings_raw = result.get("embeddings")

        # Check ids safely (may be numpy array)
        if ids is None:
            return search_results
        try:
            if len(ids) == 0:
                return search_results
        except (TypeError, ValueError):
            return search_results

        embeddings = []
        # Handle embeddings - may be None or numpy array
        if embeddings_raw is not None:
            try:
                for e in embeddings_raw:
                    if hasattr(e, "tolist") and callable(getattr(e, "tolist", None)):
                        try:
                            embeddings.append(list(cast(Any, e).tolist()))
                        except (AttributeError, TypeError):
                            embeddings.append(list(e) if isinstance(e, (list, tuple)) else [])
                    elif isinstance(e, (list, tuple)):
                        embeddings.append([float(x) for x in e if isinstance(x, (int, float))])
                    elif isinstance(e, (int, float)):
                        embeddings.append([float(e)])
                    else:
                        embeddings.append([])
            except (TypeError, ValueError):
                # numpy array iteration issue
                embeddings = []

        try:
            for idx, id_ in enumerate(ids):
                doc_metadata = dict(metadatas[idx]) if metadatas and idx < len(metadatas) and metadatas[idx] else {}
                document = documents[idx] if documents and idx < len(documents) else ""

                # Calculate simple keyword score if query provided
                if query and document:
                    query_terms = set(query.lower().split())
                    doc_terms = set(document.lower().split())
                    overlap = len(query_terms & doc_terms)
                    doc_metadata["keyword_score"] = overlap / max(len(query_terms), 1)

                # Extract the fields we added to metadata
                name_val = doc_metadata.pop("name", None)
                content_id_val = doc_metadata.pop("content_id", None)

                # Convert types to match Document constructor expectations
                name = str(name_val) if name_val is not None and not isinstance(name_val, str) else name_val
                content_id = (
                    str(content_id_val)
                    if content_id_val is not None and not isinstance(content_id_val, str)
                    else content_id_val
                )
                content = str(document) if document is not None else ""
                embedding = embeddings[idx] if idx < len(embeddings) else None

                search_results.append(
                    Document(
                        id=id_,
                        name=name,
                        meta_data=doc_metadata,
                        content=content,
                        embedding=embedding,
                        content_id=content_id,
                    )
                )
        except Exception as e:
            log_error(f"Error building get results: {e}")

        return search_results

    def _convert_filters(self, filters: Dict[str, Any]) -> Dict[str, Any]:
        """Convert simple filters to ChromaDB's filter format.

        Handles conversion of simple key-value filters to ChromaDB's operator format
        when needed.
        """
        if not filters:
            return {}

        # If filters already use ChromaDB operators ($eq, $ne, etc.), return as is
        if any(key.startswith("$") for key in filters.keys()):
            return filters

        # Convert simple key-value pairs to ChromaDB's format
        converted = {}
        for key, value in filters.items():
            if isinstance(value, (list, tuple)):
                # Convert lists to $in operator
                converted[key] = {"$in": list(value)}
            else:
                # Convert simple equality to $eq
                converted[key] = {"$eq": value}

        return converted

    async def async_search(
        self, query: str, limit: int = 5, filters: Optional[Union[Dict[str, Any], List[FilterExpr]]] = None
    ) -> List[Document]:
        """Search asynchronously by running in a thread."""
        return await asyncio.to_thread(self.search, query, limit, filters)

    def drop(self) -> None:
        """Delete the collection."""
        if self.exists():
            log_debug(f"Deleting collection: {self.collection_name}")
            self.client.delete_collection(name=self.collection_name)

    async def async_drop(self) -> None:
        """Drop the collection asynchronously by running in a thread."""
        await asyncio.to_thread(self.drop)

    def exists(self) -> bool:
        """Check if the collection exists."""
        try:
            self.client.get_collection(name=self.collection_name)
            return True
        except Exception as e:
            log_debug(f"Collection does not exist: {e}")
        return False

    async def async_exists(self) -> bool:
        """Check if collection exists asynchronously by running in a thread."""
        return await asyncio.to_thread(self.exists)

    def get_count(self) -> int:
        """Get the count of documents in the collection."""
        if self.exists():
            try:
                collection: Collection = self.client.get_collection(name=self.collection_name)
                return collection.count()
            except Exception as e:
                logger.error(f"Error getting count: {e}")
        return 0

    def optimize(self) -> None:
        raise NotImplementedError

    def delete(self) -> bool:
        try:
            self.client.delete_collection(name=self.collection_name)
            return True
        except Exception as e:
            logger.error(f"Error clearing collection: {e}")
            return False

    def delete_by_id(self, id: str) -> bool:
        """Delete document by ID."""
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Check if document exists
            if not self.id_exists(id):
                log_info(f"Document with ID '{id}' not found")
                return False

            # Delete the document
            collection.delete(ids=[id])
            log_info(f"Deleted document with ID '{id}'")
            return True
        except Exception as e:
            logger.error(f"Error deleting document by ID '{id}': {e}")
            return False

    def delete_by_name(self, name: str) -> bool:
        """Delete documents by name."""
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Find all documents with the given name
            result = collection.get(where=cast(Any, {"name": {"$eq": name}}))
            ids_to_delete = result.get("ids", [])

            if not ids_to_delete:
                log_info(f"No documents found with name '{name}'")
                return False

            # Delete all matching documents
            collection.delete(ids=ids_to_delete)
            log_info(f"Deleted {len(ids_to_delete)} documents with name '{name}'")
            return True
        except Exception as e:
            logger.error(f"Error deleting documents by name '{name}': {e}")
            return False

    def delete_by_metadata(self, metadata: Dict[str, Any]) -> bool:
        """Delete documents by metadata."""
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Build where clause for metadata filtering
            where_clause = {}
            for key, value in metadata.items():
                where_clause[key] = {"$eq": value}

            # Find all documents with the matching metadata
            result = collection.get(where=cast(Any, where_clause))
            ids_to_delete = result.get("ids", [])

            if not ids_to_delete:
                log_info(f"No documents found with metadata '{metadata}'")
                return False

            # Delete all matching documents
            collection.delete(ids=ids_to_delete)
            log_info(f"Deleted {len(ids_to_delete)} documents with metadata '{metadata}'")
            return True
        except Exception as e:
            logger.error(f"Error deleting documents by metadata '{metadata}': {e}")
            return False

    def delete_by_content_id(self, content_id: str) -> bool:
        """Delete documents by content ID."""
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Find all documents with the given content_id
            result = collection.get(where=cast(Any, {"content_id": {"$eq": content_id}}))
            ids_to_delete = result.get("ids", [])

            if not ids_to_delete:
                log_info(f"No documents found with content_id '{content_id}'")
                return False

            # Delete all matching documents
            collection.delete(ids=ids_to_delete)
            log_info(f"Deleted {len(ids_to_delete)} documents with content_id '{content_id}'")
            return True
        except Exception as e:
            logger.error(f"Error deleting documents by content_id '{content_id}': {e}")
            return False

    def _delete_by_content_hash(self, content_hash: str) -> bool:
        """Delete documents by content hash."""
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Find all documents with the given content_hash
            result = collection.get(where=cast(Any, {"content_hash": {"$eq": content_hash}}))
            ids_to_delete = result.get("ids", [])

            if not ids_to_delete:
                log_info(f"No documents found with content_hash '{content_hash}'")
                return False

            # Delete all matching documents
            collection.delete(ids=ids_to_delete)
            log_info(f"Deleted {len(ids_to_delete)} documents with content_hash '{content_hash}'")
            return True
        except Exception as e:
            logger.error(f"Error deleting documents by content_hash '{content_hash}': {e}")
            return False

    def id_exists(self, id: str) -> bool:
        """Check if a document with the given ID exists in the collection.

        Args:
            id (str): The document ID to check.

        Returns:
            bool: True if the document exists, False otherwise.
        """
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)
            # Try to get the document by ID
            result = collection.get(ids=[id])
            found_ids = result.get("ids", [])

            # Return True if the document was found
            return len(found_ids) > 0
        except Exception as e:
            logger.error(f"Error checking if ID '{id}' exists: {e}")
            return False

    def content_hash_exists(self, content_hash: str) -> bool:
        """Check if documents with the given content hash exist."""
        if not self.client:
            logger.error("Client not initialized")
            return False

        try:
            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Try to query for documents with the given content_hash
            try:
                result = collection.get(where=cast(Any, {"content_hash": {"$eq": content_hash}}))
                # Safely extract ids from result
                if hasattr(result, "get") and callable(result.get):
                    found_ids = result.get("ids", [])
                elif hasattr(result, "__getitem__") and "ids" in result:
                    found_ids = result["ids"]
                else:
                    found_ids = []

                # Return True if any documents were found
                if isinstance(found_ids, (list, tuple)):
                    return len(found_ids) > 0
                elif isinstance(found_ids, int):
                    # Some ChromaDB versions might return a count instead of a list
                    return found_ids > 0
                else:
                    return False

            except TypeError as te:
                if "object of type 'int' has no len()" in str(te):
                    # Known issue with ChromaDB 0.5.0 - internal bug
                    # As a workaround, assume content doesn't exist to allow processing to continue
                    logger.warning(
                        f"ChromaDB internal error (version 0.5.0 bug): {te}. Assuming content_hash '{content_hash}' does not exist."
                    )
                    return False
                else:
                    raise te

        except Exception as e:
            logger.error(f"Error checking if content_hash '{content_hash}' exists: {e}")
            return False

    def update_metadata(self, content_id: str, metadata: Dict[str, Any]) -> None:
        """
        Update the metadata for documents with the given content_id.

        Args:
            content_id (str): The content ID to update
            metadata (Dict[str, Any]): The metadata to update
        """
        try:
            if not self.client:
                logger.error("Client not initialized")
                return

            collection: Collection = self.client.get_collection(name=self.collection_name)

            # Find documents with the given content_id
            try:
                result = collection.get(where=cast(Any, {"content_id": {"$eq": content_id}}))

                # Extract IDs and current metadata
                if hasattr(result, "get") and callable(result.get):
                    ids = result.get("ids", [])
                    current_metadatas = result.get("metadatas", [])
                elif hasattr(result, "__getitem__"):
                    ids = result.get("ids", []) if "ids" in result else []
                    current_metadatas = result.get("metadatas", []) if "metadatas" in result else []
                else:
                    ids = []
                    current_metadatas = []

                if not ids:
                    log_debug(f"No documents found with content_id: {content_id}")
                    return

                # Flatten the new metadata first
                flattened_new_metadata = self._flatten_metadata(metadata)

                # Merge metadata for each document
                updated_metadatas = []
                for i, current_meta in enumerate(current_metadatas or []):
                    if current_meta is None:
                        meta_dict: Dict[str, Any] = {}
                    else:
                        meta_dict = dict(current_meta)  # Convert Mapping to dict

                    # Update with flattened metadata
                    meta_dict.update(flattened_new_metadata)
                    updated_metadatas.append(meta_dict)

                # Convert to the expected type for ChromaDB
                chroma_metadatas = cast(List[Mapping[str, Union[str, int, float, bool]]], updated_metadatas)
                chroma_metadatas = [{k: v for k, v in m.items() if k and v} for m in chroma_metadatas]
                collection.update(ids=ids, metadatas=chroma_metadatas)  # type: ignore
                log_debug(f"Updated metadata for {len(ids)} documents with content_id: {content_id}")

            except TypeError as te:
                if "object of type 'int' has no len()" in str(te):
                    log_warning(
                        f"ChromaDB internal error (version 0.5.0 bug): {te}. Cannot update metadata for content_id '{content_id}'."
                    )
                    return
                else:
                    raise te

        except Exception as e:
            log_error(f"Error updating metadata for content_id '{content_id}': {e}")
            raise

    def get_supported_search_types(self) -> List[str]:
        """Get the supported search types for this vector database."""
        return [SearchType.vector, SearchType.keyword, SearchType.hybrid]
